# Copyright (c) 2024 Lucas Dooms

# post-processing to find the movement of the SMC relative to the DNA

from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from io import StringIO
from itertools import islice
from pathlib import Path
from runpy import run_path
from sys import argv
from time import time
from typing import Dict, List, Tuple

import numpy as np


def timer(func):
    def timed_func(*args, **kwargs):
        time_start = time()
        ret_value = func(*args, **kwargs)
        time_end = time()
        print(time_end - time_start)
        return ret_value
    return timed_func


cached = dict()
def timer_accumulator(func):
    global cached
    if func not in cached:
        cached[func] = 0.0
    def timed_func(*args, **kwargs):
        time_start = time()
        ret_value = func(*args, **kwargs)
        time_end = time()
        cached[func] += time_end - time_start
        return ret_value
    return timed_func


@dataclass
class Box:
    xlo: float
    xhi: float
    ylo: float
    yhi: float
    zlo: float
    zhi: float

    def is_in_box(self, xyz) -> bool:
        xyz = np.array(xyz)
        condition_x = np.logical_and(self.xlo <= xyz[:,0], xyz[:,0] <= self.xhi)
        condition_y = np.logical_and(self.ylo <= xyz[:,1], xyz[:,1] <= self.yhi)
        condition_z = np.logical_and(self.zlo <= xyz[:,2], xyz[:,2] <= self.zhi)

        return np.logical_and.reduce([condition_x, condition_y, condition_z], axis=0)


class Plane:

    class Side(Enum):
        OUTSIDE = -1 # on the side of the plane that the normal vector is pointing to
        INSIDE = 1 # opposite of OUTSIDE

        @classmethod
        def get_opposite(cls, side: Plane.Side) -> Plane.Side:
            if side == cls.INSIDE:
                return cls.OUTSIDE
            elif side == cls.OUTSIDE:
                return cls.INSIDE
            raise ValueError("unknown Side value")

    def __init__(self, point: List[float], normal: List[float]):
        """point: a point on the plain,
        normal: normal vector of the plain (always normalized)"""
        normal_length = np.linalg.norm(normal)
        if normal_length == 0:
            raise ValueError("normal vector may not be zero")
        self.normal = normal / normal_length
        # take point vector to be parallel to normal vector for convenience
        # this is garantueed to still be on the same plane
        # self.point = point.dot(self.normal) * self.normal
        self.point = point

    def is_on_side(self, side: Plane.Side, point) -> bool:
        # includes points on the plane itself
        compare = self.point.dot(self.normal)
        # for checking if inside: (point - self.point) . normal <= 0
        # thus point . normal <= self.point . normal
        # for outside: the inequality is flipped, which is equivalent
        # to multiplying both sides by (-1) (without actually flipping the inequality)
        return point.dot(self.normal) * side.value <= compare * side.value


@dataclass
class LammpsData:

    ids: List[int]
    types: List[int]
    positions: List[List[float]]

    @classmethod
    def empty(cls):
        return cls(
            ids=np.array([], dtype=int),
            types=np.array([], dtype=int),
            positions=np.array([], dtype=float).reshape(-1, 3)
        )

    @timer_accumulator
    def filter(self, keep) -> None:
        """filters the current lists
        keep takes id (int), type (int), pos (array[float]) as input and returns bool"""
        keep_indices = keep(self.ids, self.types, self.positions)

        self.ids = self.ids[keep_indices]
        self.types = self.types[keep_indices]
        self.positions = self.positions[keep_indices]

    def filter_by_types(self, types: List[int]) -> None:
        self.filter(lambda _, t, __: np.isin(t, types))

    def __deepcopy__(self, memo) -> LammpsData:
        new = LammpsData(
            np.copy(self.ids),
            np.copy(self.types),
            np.copy(self.positions)
        )
        return new

    def delete_outside_box(self, box: Box) -> LammpsData:
        """creates a new LammpsData instance with points outside of the Box removed"""
        new = deepcopy(self)
        new.filter(lambda _, __, position: box.is_in_box(position))
        return new

    def delete_side_of_plane(self, plane: Plane, side: Plane.Side) -> None:
        """filters the current LammpsData instance to remove points on one side of a plane
        side: the side of the Plane that will get deleted"""
        self.filter(
            lambda _, __, pos: plane.is_on_side(Plane.Side.get_opposite(side), pos)
        )

    def combine_by_ids(self, other: LammpsData):
        """merges two LammpsData instances by keeping all values present in any of the two
        mutates the self argument"""
        # WARNING: making no guarantees about the order
        all_ids = np.concatenate([self.ids, other.ids])
        all_types = np.concatenate([self.types, other.types])
        all_positions = np.concatenate([self.positions, other.positions])
        self.ids, indices = np.unique(all_ids, return_index=True)
        self.types = all_types[indices]
        self.positions = all_positions[indices]

    def get_position_from_index(self, index):
        return self.positions[np.where(index == self.ids)[0][0]]


class Parser:

    atom_format = "ITEM: ATOMS id type x y z\n"

    class EndOfLammpsFile(Exception):
        pass

    def __init__(self, file_name: str) -> None:
        self.file = open(file_name, 'r', encoding='utf-8')

    def skip_to_atoms(self) -> Dict[str, str]:
        saved = dict()
        current_line = None
        empty = True
        for line in self.file:
            empty = False
            if line.startswith("ITEM: ATOMS"):
                if line != self.atom_format:
                    raise ValueError(f"Wrong format of atoms, found\n{line}\nshould be\n{self.atom_format}\n")
                return saved
            # remove newline
            line = line[:-1]
            if line.startswith("ITEM:"):
                saved[line] = []
                current_line = line
            else:
                saved[current_line].append(line)
        if empty:
            raise self.EndOfLammpsFile()
        raise ValueError("reached end of file unexpectedly")

    @staticmethod
    @timer_accumulator
    def get_array(lines):
        lines = "".join(lines)
        with StringIO(lines) as file:
            array = np.loadtxt(file)
        return array

    def next_step(self) -> Tuple[int, List[List[float]]]:
        """returns timestep and list of [id, type, x, y, z] for each atom"""

        saved = self.skip_to_atoms()
        timestep = int(saved["ITEM: TIMESTEP"][0])
        number_of_atoms = int(saved["ITEM: NUMBER OF ATOMS"][0])

        lines = list(islice(self.file, number_of_atoms))
        if len(lines) != number_of_atoms:
            raise ValueError("reached end of file unexpectedly")

        array = self.get_array(lines)

        return timestep, array

    @staticmethod
    def split_data(array) -> LammpsData:
        """split array into ids, types, xyz"""
        ids, types, x, y, z = array.transpose()
        xyz = np.concatenate([x, y, z]).reshape(3, -1).transpose()
        return LammpsData(np.array(ids, dtype=int), np.array(types, dtype=int), xyz)

    def __del__(self) -> None:
        self.file.close()


def create_box(data: LammpsData, types: List[int]) -> Box:
    copy_data = deepcopy(data)
    copy_data.filter_by_types(types)
    reduced_xyz = copy_data.positions

    return Box(
        xlo = np.min(reduced_xyz[:,0]),
        xhi = np.max(reduced_xyz[:,0]),
        ylo = np.min(reduced_xyz[:,1]),
        yhi = np.max(reduced_xyz[:,1]),
        zlo = np.min(reduced_xyz[:,2]),
        zhi = np.max(reduced_xyz[:,2])
    )


def distance_point_to_plane(point, point_on_plane, normal_direction) -> float:
    return abs((point - point_on_plane).dot(normal_direction))


def get_normal_direction(p1, p2, p3):
    perpendicular = np.cross(p1 - p2, p1 - p3)
    return perpendicular / np.linalg.norm(perpendicular)


def write(file, steps, positions):
    for step, position in zip(steps, positions):
        file.write("ITEM: TIMESTEP\n")
        file.write(f"{step}\n")
        file.write("ITEM: NUMBER OF ATOMS\n")
        file.write("1\n")
        file.write(
            "ITEM: BOX BOUNDS ff ff ff\n"
            "-8.5170000000000005e+02 8.5170000000000005e+02\n"
            "-8.5170000000000005e+02 8.5170000000000005e+02\n"
            "-8.5170000000000005e+02 8.5170000000000005e+02\n"
        )
        file.write(Parser.atom_format)
        file.write(f"1 1 {position[0]} {position[1]} {position[2]}\n")


def is_sorted(lst) -> bool:
    return all(lst[i] <= lst[i + 1] for i in range(len(lst) - 1))


def split_into_index_groups(indices) -> List[List[int]]:
    """splits a (sorted) list of integers into groups of adjacent integers"""
    indices = list(indices)
    if not indices:
        return indices

    if not is_sorted(indices):
        raise ValueError("indices must be sorted")
    
    groups = [[indices[0]]]
    for index in indices[1:]:
        if groups[-1][-1] == index - 1:
            groups[-1].append(index)
        else:
            groups.append([index])

    return groups


def get_bead_distances(positions, pos1: int, pos2: int, pos3: int):
    normal_vector = get_normal_direction(pos1, pos2, pos3)

    return distance_point_to_plane(positions, pos1, normal_vector)


def remove_outside_planar_n_gon(data: LammpsData, points, delta_out_of_plane: float, delta_extended_plane: float):
    """removes all points outside a box created from an n-sided polygon in a plane.
    the points are the vertices of the n-gon and are assumed to be in the same plane.
    delta is the thickness of the box in each direction going out of the plane."""
    if len(points) < 3:
        raise ValueError("there must be at least 3 points in the n-gon")
    # normal to n-gon plane
    try:
        normal = np.cross(points[1] - points[0], points[10] - points[9])
    except IndexError:
        normal = np.cross(points[1] - points[0], points[2] - points[1])

    # normalize the normal! (will be used to apply thickness delta)
    normal = normal / np.linalg.norm(normal)

    # delete points further than delta perpendicular to the n-gon
    plane = Plane(points[0] + delta_out_of_plane * normal, normal)
    data.delete_side_of_plane(plane, Plane.Side.OUTSIDE)
    plane = Plane(points[0] - delta_out_of_plane * normal, normal)
    data.delete_side_of_plane(plane, Plane.Side.INSIDE)

    # delete points far away in the plane of the n-gon
    for point1, point2 in zip(points, points[1:] + [points[0]]):
        side_vector = point2 - point1
        normal_to_side = np.cross(normal, side_vector) # always points INSIDE
        normal_to_side = normal_to_side / np.linalg.norm(normal_to_side)
        side_plane = Plane(point1 - delta_extended_plane * normal_to_side, normal_to_side)
        # INSIDE of plane points out of the shape
        data.delete_side_of_plane(side_plane, Plane.Side.INSIDE)


def handle_dna_bead(data: LammpsData, new_data: LammpsData, indices, positions, parameters, step):
    if len(new_data.positions) == 0:
        indices.append(-1)
        positions.append(data.positions[indices[-1]])
        return

    pos_top_left = data.get_position_from_index(parameters["top_left_bead_id"])
    pos_top_right = data.get_position_from_index(parameters["top_right_bead_id"])
    pos_left = data.get_position_from_index(parameters["left_bead_id"])
    pos_right = data.get_position_from_index(parameters["right_bead_id"])
    pos_middle_left = data.get_position_from_index(parameters["middle_left_bead_id"])
    pos_middle_right = data.get_position_from_index(parameters["middle_right_bead_id"])
    pos_kleisins = [data.get_position_from_index(i) for i in parameters["kleisin_ids"]]

    delta = 0.51 * parameters["dna_spacing"]
    small_delta = delta / 4.0

    new_data_copy1 = deepcopy(new_data)
    remove_outside_planar_n_gon(new_data_copy1, [pos_top_left, pos_left, pos_right], delta, small_delta)

    new_data_copy2 = deepcopy(new_data)
    remove_outside_planar_n_gon(new_data_copy2, [pos_top_left, pos_top_right, pos_right], delta, small_delta)

    new_data_copy3 = deepcopy(new_data)
    remove_outside_planar_n_gon(new_data_copy3, [pos_middle_right, pos_middle_left, pos_left], delta, small_delta)

    new_data_copy4 = deepcopy(new_data)
    remove_outside_planar_n_gon(new_data_copy4, [pos_middle_right, pos_left, pos_right], delta, small_delta)

    new_data_copy5 = deepcopy(new_data)
    remove_outside_planar_n_gon(new_data_copy5, pos_kleisins, delta, small_delta)

    new_data = LammpsData.empty()
    new_data.combine_by_ids(new_data_copy1)
    new_data.combine_by_ids(new_data_copy2)
    new_data.combine_by_ids(new_data_copy3)
    new_data.combine_by_ids(new_data_copy4)
    new_data.combine_by_ids(new_data_copy5)

    if len(new_data.positions) == 0:
        print(f"call: {step}")
        indices.append(-1)
        positions.append(data.positions[indices[-1]])
        return

    # find groups
    grps = split_into_index_groups(new_data.ids)

    grp = grps[0]
    grp = [np.where(new_data.ids == id)[0][0] for id in grp]

    # alternative distance method
    # distances = get_bead_distances(new_data.positions, pos_top, pos_left, pos_right)
    # closest_val = np.min(distances[grp])
    # closest_bead_index = np.where(distances == closest_val)[0][0]

    # TODO: assumes lowest index gives the desired bead
    closest_bead_index = grp[0]

    indices.append(new_data.ids[closest_bead_index])
    positions.append(new_data.positions[closest_bead_index])


def get_best_match_dna_bead_in_smc(folder_path):
    """
    For each timestep:
        create box around SMC
        remove DNA not within box
        remove DNA part of lower segment (we only want the upper segment here)
        find DNA closest to SMC plane
    """
    parameters = run_path((path / "post_processing_parameters.py").as_posix())

    par = Parser(folder_path / "output.lammpstrj")
    dna_indices_list = parameters["dna_indices_list"]
    steps = []
    indices_array = [[] for _ in range(len(dna_indices_list))]
    positions_array = [[] for _ in range(len(dna_indices_list))]
    while True:
        try:
            step, arr = par.next_step()
        except Parser.EndOfLammpsFile:
            break

        steps.append(step)

        data = Parser.split_data(arr)
        # TODO: get range from post_processing_parameters.py
        box = create_box(data, parameters["SMC_types"])

        new_data = data.delete_outside_box(box)
        new_data.filter_by_types(parameters["DNA_types"])
        # split, and call for each
        for i, (min_index, max_index) in enumerate(dna_indices_list):
            new_data_temp = deepcopy(new_data)
            new_data_temp.filter(lambda id, _, __: np.logical_and(min_index <= id, id <= max_index))
            handle_dna_bead(data, new_data_temp, indices_array[i], positions_array[i], parameters, step)

    # delete old files
    for p in folder_path.glob("marked_bead*.lammpstrj"):
        p.unlink()

    for i, positions in enumerate(positions_array):
        with open(folder_path / f"marked_bead{i}.lammpstrj", 'w', encoding='utf-8') as file:
            write(file, steps, positions)

    print(cached)

    for i, indices in enumerate(indices_array):
        np.savez(folder_path / f"bead_indices{i}.npz", steps=steps, ids=indices)

    import matplotlib.pyplot as plt

    plt.figure(figsize=(8, 6), dpi=144)
    plt.title("Index of DNA bead inside SMC loop in time")
    plt.xlabel("time")
    plt.ylabel("DNA bead index")
    for i in range(len(indices_array)):
        plt.scatter(steps, indices_array[i], s=0.5, label=f"DNA {i}")
    plt.legend()
    plt.savefig(folder_path / "bead_id_in_time.png")


def get_msd_obstacle(folder_path):
    par = Parser(folder_path / "obstacle.lammpstrj")
    steps = []
    positions = []
    while True:
        try:
            step, arr = par.next_step()
        except Parser.EndOfLammpsFile:
            break

        steps.append(step)
        positions.append(arr[2])

    print(cached)

    # calculate msd in time chunks
    time_chunk_size = 2000 # number of timesteps to pick for one window

    def calculate_msd(array) -> float:
        return np.average((array - array[0])**2)

    def apply_moving_window(window_size: int, func, array):
        """returns the array of the results from applying the func to
        windows along the array."""
        result = []
        for i in range(len(array) - window_size + 1):
            result.append(
                func(array[i:i + window_size])
            )
        return result

    msd_array = apply_moving_window(time_chunk_size, calculate_msd, positions)

    import matplotlib.pyplot as plt

    plt.figure(figsize=(8, 6), dpi=144)
    plt.title(f"MSD over time in chunks of {(steps[1] - steps[0]) * time_chunk_size} (all average = {calculate_msd(positions)})")
    plt.xlabel("time")
    plt.ylabel("MSD")
    plt.scatter(steps[:-time_chunk_size+1], msd_array, s=0.3)
    plt.savefig(folder_path / "msd_in_time.png")


def test_plane_distances():
    p1 = np.array([0, 1, 0], dtype=float)
    p2 = np.array([0, 1, 2], dtype=float)
    p3 = np.array([1, 0, 1], dtype=float)
    n = get_normal_direction(p1, p2, p3)
    print(n)
    point = np.array([2, 2, 5], dtype=float)
    another_one = np.array([1, 1 ,1], dtype=float)
    print(distance_point_to_plane(np.array([point, another_one, another_one]), p1, n))


def test_plane_comparisons():
    point_on_plane = np.array([0, 0, 0], dtype=float)
    n = np.array([1, 0, 0], dtype=float)
    plane = Plane(point_on_plane, n)
    points = np.array([[0, 0, 0], [0.5, 0, 0], [-20, 0, 0]])
    print(plane.is_on_side(Plane.Side.INSIDE, points)) # should be [True, False, True]
    print(plane.is_on_side(Plane.Side.OUTSIDE, points)) # should be [True, True, False]

    plane2 = Plane(np.array([1, 0, 0], dtype=float), np.array([1, 1, 0], dtype=float))
    print(plane2.is_on_side(Plane.Side.INSIDE, points)) # should be [True, True, True]
    print(plane2.is_on_side(Plane.Side.INSIDE, np.array([1, 0.2, 0], dtype=float))) # should be False


if __name__ == "__main__":
    argv = argv[1:]
    if len(argv) != 1:
        raise Exception("Please provide a folder path")
    path = Path(argv[0])
    get_best_match_dna_bead_in_smc(path)
    get_msd_obstacle(path)



from pathlib import Path
from runpy import run_path
from sys import argv

import numpy as np


def read_lammpstrj(file_path):
    timesteps = []
    num_atoms = []
    box_bounds = []
    atom_data = []

    with open(file_path, "r") as file:
        lines = file.readlines()

    i = 0

    while i < len(lines):
        if lines[i].strip() == "ITEM: TIMESTEP":
            timestep = int(lines[i + 1].strip())

            timesteps.append(timestep)

            i += 2

        elif lines[i].strip() == "ITEM: NUMBER OF ATOMS":
            n_atoms = int(lines[i + 1].strip())

            num_atoms.append(n_atoms)

            i += 2

        elif lines[i].strip() == "ITEM: BOX BOUNDS ff ff ff":
            bounds = [
                list(map(float, lines[i + j].strip().split())) for j in range(1, 4)
            ]

            box_bounds.append(bounds)

            i += 4

        elif lines[i].strip() == "ITEM: ATOMS id type x y z":
            atoms = []

            for j in range(num_atoms[-1]):
                atoms.append(list(map(float, lines[i + 1 + j].strip().split())))

            atom_data.append(atoms)

            i += 1 + num_atoms[-1]

        else:
            i += 1

    return timesteps, num_atoms, box_bounds, atom_data


def write_lammpstrj(file_path, timesteps, num_atoms, box_bounds, atom_data):
    with open(file_path, "w") as file:
        for t, n, bounds, atoms in zip(timesteps, num_atoms, box_bounds, atom_data):
            file.write("ITEM: TIMESTEP\n")

            file.write(f"{t}\n")

            file.write("ITEM: NUMBER OF ATOMS\n")

            file.write(f"{n}\n")

            file.write("ITEM: BOX BOUNDS ff ff ff\n")

            for b in bounds:
                file.write(f"{b[0]} {b[1]}\n")

            file.write("ITEM: ATOMS id type x y z\n")

            for atom in atoms:
                file.write(" ".join(map(str, atom)) + "\n")


def rigid_transform_3D(A, B):
    """implementation of Kabsch algorithm"""
    assert len(A) == len(B)

    centroid_A = np.mean(A, axis=0)
    centroid_B = np.mean(B, axis=0)

    # Center the points
    AA = A - centroid_A
    BB = B - centroid_B

    H = AA.transpose().dot(BB)
    U, _, Vt = np.linalg.svd(H)
    R = U.dot(Vt)

    # Special reflection case
    if np.linalg.det(R) < 0:
        Vt[-1, :] *= -1
        R = U.dot(Vt)

    t = centroid_B - R.transpose().dot(centroid_A)

    return R.transpose(), t.transpose()


def transform_atoms(atom_data, index1, index2, index3, v0, v1, v2):
    for timestep_data in atom_data:
        A = np.array(
            [
                timestep_data[index1 - 1][2:5],
                timestep_data[index2 - 1][2:5],
                timestep_data[index3 - 1][2:5],
            ]
        )

        B = np.array([v0, v1, v2])

        R, t = rigid_transform_3D(A, B)

        for atom in timestep_data:
            atom_pos = np.array(atom[2:5])

            transformed_pos = np.dot(R, atom_pos) + t

            atom[2:5] = transformed_pos.tolist()

    return atom_data


def main(input_file, output_file, index1, index2, index3):
    timesteps, num_atoms, box_bounds, atom_data = read_lammpstrj(input_file)

    # Extract initial positions for the three atoms
    initial_pos1 = atom_data[0][index1 - 1][2:5]
    initial_pos2 = atom_data[0][index2 - 1][2:5]
    initial_pos3 = atom_data[0][index3 - 1][2:5]

    # Transform the atom positions to keep index1, index2, and index3 at their initial positions
    transformed_atom_data = transform_atoms(
        atom_data, index1, index2, index3, initial_pos1, initial_pos2, initial_pos3
    )

    write_lammpstrj(
        output_file, timesteps, num_atoms, box_bounds, transformed_atom_data
    )


if __name__ == "__main__":
    argv = argv[1:]
    if len(argv) != 2:
        raise ValueError(
            "2 inputs required: output.lammpstrj and post_processing_parameters.py"
        )

    output_file = Path(argv[0])

    post_processing_parameters_file = Path(argv[1])
    parameters = run_path(post_processing_parameters_file.as_posix())
    kleisin_ids = parameters["kleisin_ids"]
    ref_ids = [kleisin_ids[1], kleisin_ids[len(kleisin_ids) // 2], kleisin_ids[-2]]

    main(output_file, output_file.parent / f"perspective.{output_file.name}", *ref_ids)



# Copyright (c) 2024 Lucas Dooms

import math

import numpy as np

from smc_lammps.generate.structures.structure_creator import (
    attach,
    attach_chain,
    get_circle_segment,
    get_interpolated,
    get_straight_segment,
)
from smc_lammps.generate.util import get_closest


def get_dna_coordinates_straight(nDNA: int, DNAbondLength: float):
    rDNA = get_straight_segment(nDNA, [1, 0, 0])

    # Shift X-coordinate to get DNA end-point at X = 0

    rDNA[:, 0] -= rDNA[0][0]

    # get correct bead spacings

    rDNA *= DNAbondLength

    # Rotate (flip the x-component)

    rDNA[:, 0] *= -1

    return [rDNA]


def get_dna_coordinates_safety_belt(nDNA: int, DNAbondLength: float):
    rDNA = get_interpolated(
        DNAbondLength,
        10
        * DNAbondLength
        * np.array(
            [
                [4, 0, 0],
                [2, 0, 0],  # right, straight piece
                [0, 1.0, 0],
                [-0.5, 1.0, 0],
                [-3, 1.25, 0],  # up, to the left
                [-3.25, 0.75, 0],  # down, to the left
                [-0.5, -1, 0],
                [0, -1, 0],
                [0.7, -1, 0],
                [1.5, -1.5, 0],  # down, to the right
                [1.0, -2, 0],
                [-2, -2, 0],  # down, to the left
                [-3, -0.5, 0],
                [-4, 0, 0],  # left, straight
            ],
            dtype=float,
        ),
    )
    distances = np.linalg.norm(
        rDNA - 10 * DNAbondLength * np.array([0, -1.0, 0.0]), axis=1
    )
    belt_index = np.where(distances == np.min(distances))[0][0]

    remaining = nDNA - len(rDNA)
    nLeft = int(remaining * 0.05)
    nRight = remaining - nLeft
    left = get_straight_segment(nLeft + 1, [-1, 0, 0]) * DNAbondLength
    right = get_straight_segment(nRight + 1, [-1, 0, 0]) * DNAbondLength
    right, rDNA, left = attach_chain(right, [[rDNA, True], [left, True]])

    belt_location = rDNA[belt_index]
    return (
        [np.concatenate([right, rDNA, left])],
        belt_location,
        belt_index + len(right),
        len(right) - 40,
    )


def get_dna_coordinates_advanced_safety_belt(nDNA: int, DNAbondLength: float):
    smc_pos = np.array([3.3, -2.1, 0.0])
    rDNA = get_interpolated(
        DNAbondLength,
        10
        * DNAbondLength
        * np.array(
            [
                # right, straight piece
                [6, -2.0, 0],
                [5, -1.4, 0],
                [4, 0, 0],
                [2, 0, 0],
                # right, straight piece
                # up, to the left
                [0, 1.0, 0],
                [-3, 1.5, 0],
                [-4, 1.25, 0],
                # up, to the left
                # down, to the left
                [-5, 0.75, 0],
                [-6, 0.5, 0],
                [-7, 0, 0],
                # down, to the left
                # down, to the right
                [-5.5, -1, 0],
                [-4, -1.3, 0],
                [0.7, -1.4, 0],
                [2.8, -1.8, 0],
                [3.5, -2.2, 0.0],
                [3.5, -2.5, 0.0],
                # down, to the right
                # down, to the left
                [3, -3.1, 0],
                [2.5, -3.2, 0],
                [2, -3.2, 0],
                # down, to the left
            ],
            dtype=float,
        ),
    )
    distances = np.linalg.norm(rDNA - 10 * DNAbondLength * smc_pos, axis=1)
    belt_index = np.where(distances == np.min(distances))[0][0]

    remaining = nDNA - len(rDNA)
    right = get_straight_segment(remaining + 1, [-1, 0, 0]) * DNAbondLength
    rDNA = attach(right, rDNA, delete_overlap=True)

    belt_location = rDNA[belt_index]
    bead_to_tether_id = len(right) + 31

    return (
        [np.concatenate([right, rDNA])],
        belt_location,
        belt_index + len(right),
        bead_to_tether_id,
    )


def get_dna_coordinates_advanced_safety_belt_plus_loop(nDNA: int, DNAbondLength: float):
    smc_pos = np.array([3.3, -2.1, 0.0])
    tether_pos = np.array([3.75, 0.0, 0.0])
    rDNA = get_interpolated(
        DNAbondLength,
        10
        * DNAbondLength
        * np.array(
            [
                # right piece
                [6, -2.0, 0],
                [5, -1.4, 0],
                [4.7, -0.5, 0],
                # right piece
                # loop
                [4.6, -0.4, 2],
                [4.7, -0.2, 4],
                [4.5, -0.2, 4],
                [4.0, -0.15, 4],
                [0.0, -0.15, 4],
                [1.0, -0.15, 2],
                [4.0, -0.05, 2.4],
                [4.2, 0, 2],
                # loop
                # up, to the left
                [4, 0, 0],
                [2, 0, 0],
                [0, 1.0, 0],
                [-3, 1.5, 0],
                [-4, 1.25, 0],
                # up, to the left
                # down, to the left
                [-5, 0.75, 0],
                [-6, 0.5, 0],
                [-7, 0, 0],
                # down, to the left
                # down, to the right
                [-5.5, -1, 0],
                [-4, -1.3, 0],
                [0.7, -1.4, 0],
                [2.8, -1.8, 0],
                [3.5, -2.2, 0.0],
                [3.5, -2.5, 0.0],
                # down, to the right
                # down, to the left
                [3, -3.1, 0],
                [2.5, -3.2, 0],
                [2, -3.2, 0],
                # down, to the left
                # straight left
                [-1, -3.2, 0],
                [-7, -3.2, 0],
                # straight left
            ],
            dtype=float,
        ),
    )
    belt_index = get_closest(rDNA, 10 * DNAbondLength * smc_pos)
    bead_to_tether_id = get_closest(rDNA, 10 * DNAbondLength * tether_pos)

    remaining = nDNA - len(rDNA)
    right = get_straight_segment(remaining + 1, [-1, 0, 0]) * DNAbondLength
    rDNA = attach(right, rDNA, delete_overlap=True)

    belt_location = rDNA[belt_index]
    bead_to_tether_id += len(right)

    return (
        [np.concatenate([right, rDNA])],
        belt_location,
        belt_index + len(right),
        bead_to_tether_id,
    )


def get_dna_coordinates(
    nDNA: int, DNAbondLength: float, diameter: float, nArcStraight: int
):
    # form vertical + quarter circle + straight + semi circle + horizontal parts

    # Number of beads forming the arced DNA piece (err on the high side)
    nArcedDNA = math.ceil(
        3 / 4 * math.pi * diameter / DNAbondLength
    )  # 3 / 4 = 1 / 2 + 1 / 4 = semi + quarter

    # We want an odd number (necessary for angle/dihedral interactions)
    if nArcedDNA % 2 == 0:
        nArcedDNA += 1

    # Upper DNA piece

    nUpperDNA = (nDNA - nArcedDNA - nArcStraight) // 2

    rUpperDNA = get_straight_segment(nUpperDNA, [0, -1, 0])

    # Arced DNA piece

    nArcSemi = int(nArcedDNA * 2 / 3)
    nArcQuart = nArcedDNA - nArcSemi

    # since there will be overlap: use one extra, then delete it later (after pieces are assembled)
    rArcQuart = get_circle_segment(
        nArcQuart + 1, end_inclusive=True, theta_start=0, theta_end=-np.pi / 2.0
    )

    rArcStraight = get_straight_segment(nArcStraight, [-1, 0, 0])

    rArcSemi = get_circle_segment(
        nArcSemi + 1,
        end_inclusive=True,
        theta_start=np.pi / 2.0,
        theta_end=np.pi * 3.0 / 2.0,
    )

    # Lower DNA piece

    nLowerDNA = nDNA - nUpperDNA - nArcedDNA - nArcStraight

    rLowerDNA = get_straight_segment(nLowerDNA, [1, 0, 0])

    # Total DNA

    rUpperDNA, rArcQuart, rArcStraight, rArcSemi, rLowerDNA = attach_chain(
        rUpperDNA,
        [
            [rArcQuart, True],
            [rArcStraight, False, 1.0],
            [rArcSemi, True],
            [rLowerDNA, False, 1.0],
        ],
    )

    # alternative, without attach_chain method:
    # rArcQuart = attach(rUpperDNA, rArcQuart, delete_overlap=True)
    # rArcStraight = attach(rArcQuart, rArcStraight, delete_overlap=False, extra_distance=1.0)
    # rArcSemi = attach(rArcStraight, rArcSemi, delete_overlap=True)
    # rLowerDNA = attach(rArcSemi, rLowerDNA, delete_overlap=False, extra_distance=1.0)

    rDNA = np.concatenate([rUpperDNA, rArcQuart, rArcStraight, rArcSemi, rLowerDNA])

    # Shift X-coordinate to get DNA end-point at X = 0

    rDNA[:, 0] -= rDNA[0][0]

    # get correct bead spacings

    rDNA *= DNAbondLength

    # Rotate (flip the x-component)

    rDNA[:, 0] *= -1

    # the position in the center of the semi arc
    offset = nUpperDNA + nArcQuart + nArcStraight
    centerCoordinate = rDNA[offset] + (rDNA[offset + nArcSemi] - rDNA[offset]) / 2.0

    return [rDNA], centerCoordinate


def get_dna_coordinates_twist(nDNA: int, DNAbondLength: float, diameter: float):
    # form upper + semi circle + horizontal parts

    # Number of beads forming the arced DNA piece (err on the high side)
    nArcedDNA = math.ceil(
        1 / 2 * math.pi * diameter / DNAbondLength
    )  # 1 / 2 = semi circle

    # We want an odd number (necessary for angle/dihedral interactions)
    if nArcedDNA % 2 == 0:
        nArcedDNA += 1

    # Upper DNA piece

    nUpperDNA = (nDNA - nArcedDNA) // 2

    rUpperDNA = get_straight_segment(nUpperDNA, [-1, 0, 0])

    # Arced DNA piece

    nArcSemi = nArcedDNA

    # since there will be overlap: use one extra, then delete it later (after pieces are assembled)
    rArcSemi = get_circle_segment(
        nArcSemi + 1,
        end_inclusive=True,
        theta_start=np.pi / 2.0,
        theta_end=np.pi * 3.0 / 2.0,
    )

    # Lower DNA piece

    nLowerDNA = nDNA - nUpperDNA - nArcedDNA

    rLowerDNA = get_straight_segment(nLowerDNA, [1, 0, 0])

    # Total DNA

    rUpperDNA, rArcSemi, rLowerDNA = attach_chain(
        rUpperDNA, [[rArcSemi, True], [rLowerDNA, False, 1.0]]
    )

    rDNA = np.concatenate([rUpperDNA, rArcSemi, rLowerDNA])

    # Shift X-coordinate to get DNA end-point at X = 0

    rDNA[:, 0] -= rDNA[0][0]

    # get correct bead spacings

    rDNA *= DNAbondLength

    # the position in the center of the arc
    offset = nUpperDNA
    centerCoordinate = rDNA[offset] + (rDNA[offset + nArcSemi] - rDNA[offset]) / 2.0

    return [rDNA], centerCoordinate


def get_dna_coordinates_doubled(nDNA: int, DNAbondLength: float, diameter: float):
    nOuterDNA = math.ceil(nDNA / 1.9)
    nInnerDNA = nDNA - nOuterDNA
    [rOuterDNA], outerCenter = get_dna_coordinates_twist(
        nOuterDNA, DNAbondLength, diameter
    )
    [rInnerDNA], innerCenter = get_dna_coordinates_twist(
        nInnerDNA, DNAbondLength, diameter / 1.5
    )
    # align pieces
    rInnerDNA += outerCenter - innerCenter
    # create separation (x direction)
    rInnerDNA[:, 0] += 5.0 * DNAbondLength
    return [rOuterDNA, rInnerDNA], outerCenter



from dataclasses import dataclass
from typing import Any


@dataclass
class Parameters:
    def __setattr__(self, name: str, value: Any, /) -> None:
        if not hasattr(self, name):
            raise AttributeError("You cannot define new parameters.")
        super().__setattr__(name, value)

    ################ General parameters ################

    loop = 100
    "Initial loop size (DNA beads)"

    diameter = 20
    "Diameter of initial loop (nm)"

    T = 300.0
    "Simulation temperature (K)"

    kB = 0.013806504
    "Boltzmann's constant (pN nm / K)"

    gamma = 0.5
    "Inverse of friction coefficient (ns)"

    output_steps = 10000
    "Printing period (time steps)"

    timestep = 2e-4
    "Simulation timestep (ns)"

    seed = 123
    "Random number seed"

    N: int = 501
    "Number of DNA beads"

    n = 5
    "Number of base pairs per DNA bead"

    force = 0.800
    """Stretching forces (pN) (set to any falsy value for no forces)
    WARNING: currently: if no forces -> ends are frozen"""

    runs = 10
    "Number of independent runs"

    cycles: int | None = 2
    """Number of SMC cycles (if set to None, will find approximate value using max_steps)
    Note: cycles are stochastic, so time per cycle is variable"""

    max_steps: int | None = None
    """Max steps for run (None -> no maximum, will complete every cycle)
    Note: this is not a hard limit, some extra steps may be performed to complete a cycle"""

    steps_APO = 2000000
    "Average number of steps spent in APO state (waiting for ATP binding)"

    steps_ATP = 8000000
    "Average number of steps spent in ATP state (waiting for ATP hydrolysis)"

    steps_ADP = 2000000
    "Average number of steps spent in ADP state (waiting for return to APO)"

    ##################### DNA #######################

    dna_config = "folded"
    "configuration to generate"

    add_stopper_bead = False
    "add a bead that prevents the SMC from slipping off of the wrong end of the DNA"

    add_RNA_polymerase = True
    "adds 10 nm bead at DNA-tether site"

    RNA_polymerase_size: float = 5.0
    "radius of RNA polymerase (nm)"

    RNA_polymerase_type = 1

    spaced_beads_interval: int | None = None
    "how many DNA beads to leave between small obstacles"

    spaced_beads_size: float = 5.0
    "radius of beads along DNA (nm)"

    ##################### Geometry #####################

    arm_length = 50.0
    "Length of each coiled-coil arm (nm)"

    bridge_width = 7.5
    "Width of ATP bridge (nm)"

    hinge_radius = 1.5
    "Hinge radius (nm)"

    rigid_hinge = True

    kleisin_radius = 7.0
    "Radius of lower circular-arc compartment (nm)"

    sigma_SMC_DNA = 2.5
    "SMC-DNA hard-core repulsion radius = LJ sigma (nm)"

    sigma = 2.5

    folding_angle_APO = 45.0
    "Folding angle of lower compartment (degrees)"

    folding_angle_ATP = 160.0
    "Folding angle of lower compartment (degrees)"

    arms_angle_ATP = 130.0
    "Opening angle of arms in ATP-bound state (degrees)"

    #################### LJ energies ###################

    # 3 = Repulsion
    # 4 = Upper site
    # 5 = Middle site
    # 6 = Lower site

    # LJ energy (kT units)
    # DE for a cutoff of 3.0 nm: 0.11e
    # DE for a cutoff of 3.5 nm: 0.54e

    epsilon3 = 3.0
    epsilon4 = 6.0
    epsilon5 = 6.0
    epsilon6 = 100.0

    # LJ cutoff (nm)
    cutoff4 = 3.5
    cutoff5 = 3.5
    cutoff6 = 3.0

    ################# Bending energies #################

    arms_stiffness = 100.0
    "Bending stiffness of arm-bridge angle (kT units)"

    elbows_stiffness = 30.0
    "Bending stiffness of elbows (kT units)"

    site_stiffness = 100.0
    "Alignment stiffness of binding sites (kT units)"

    folding_stiffness = 60.0
    "Folding stiffness of lower compartment (kT units)"

    asymmetry_stiffness = 100.0
    "Folding asymmetry stiffness of lower compartment (kT units)"

    ################# Other #################

    smc_force = 0.0
    "Extra force on SMC in the -x direction and +y direction (left & up)"

    use_charges = False
    "Enable Coulomb interactions in LAMMPS"



# Copyright (c) 2024 Lucas Dooms

import argparse
import subprocess
from pathlib import Path
from runpy import run_path
from typing import List, Tuple

# TODO: VMD uses zero-indexed arrays!

parser = argparse.ArgumentParser(
    prog='Visualize with VMD',
    description='Creates appropriate vmd.init file and runs vmd',
    epilog='End'
)


parser.add_argument('directory', help='the directory containing LAMMPS output files')
fn_arg = parser.add_argument('-f', '--file_name', help='name of file, default: \'output.lammpstrj\'', default='output.lammpstrj')

args = parser.parse_args()
path = Path(args.directory)

parameters = run_path((path / "post_processing_parameters.py").as_posix())

class Molecules:

    nice_color_ids = [
        7, # green
        1, # red
        9, # pink
        6, # silver
    ]

    def __init__(self, path_to_vmd_init: Path) -> None:
        self.index = -1
        self.rep_index = 0
        self.color_index = 0
        self.path = path_to_vmd_init / "vmd.init"
        # clear the file
        with open(self.path, 'w', encoding='utf-8'):
            pass

    def get_color_id(self) -> int:
        color_id = self.nice_color_ids[self.color_index % len(self.nice_color_ids)]
        self.color_index += 1
        return color_id

    def create_new(self, file_name: str, other_args: str) -> None:
        with open(self.path, 'a', encoding="utf-8") as file:
            file.write(f"mol new {file_name} {other_args}\n")
            self.index += 1

    def create_new_marked(self, file_name: str) -> None:
        self.create_new(file_name, "waitfor all")
        with open(self.path, 'a', encoding="utf-8") as file:
            file.write(f"mol modstyle 0 {self.index} vdw\n")

    def create_new_dna(self, file_name: str, dna_pieces: List[Tuple[int, int]], remove_ranges: List[Tuple[int, int]]) -> None:
        self.create_new(file_name, "waitfor all")
        with open(self.path, 'a', encoding="utf-8") as file:
            # show everything, slightly smaller
            file.write(f"mol modstyle 0 {self.index} cpk 1.3\n")

            # remove from ranges
            selections = []
            for rng in remove_ranges:
                selections.append(f"index < {rng[0] - 1} or index > {rng[1] - 1}")
            file.write(f"mol modselect 0 {self.index} " + " and ".join(selections) + "\n")

            self.add_dna_pieces(file, dna_pieces)

    def add_dna_pieces(self, file, dna_pieces: List[Tuple[int, int]]) -> None:
        # color the pieces differently
        file.write("mol rep cpk\n")
        for piece in dna_pieces:
            file.write(f"mol addrep {self.index}\n")
            self.rep_index += 1
            file.write(f"mol modselect {self.rep_index} {self.index} index >= {piece[0] - 1} and index <= {piece[1] - 1}\n")
            file.write(f"mol modcolor {self.rep_index} {self.index} colorID {self.get_color_id()}\n")
            file.write(f"mol modstyle {self.rep_index} {self.index} cpk 1.4\n")

    def add_piece(self, rng: Tuple[int, int]) -> None:
        with open(self.path, 'a', encoding="utf-8") as file:
            file.write("mol rep cpk\n")
            file.write(f"mol addrep {self.index}\n")
            self.rep_index += 1
            file.write(f"mol modselect {self.rep_index} {self.index} index >= {rng[0] - 1} and index <= {rng[1] - 1}\n")
            file.write(f"mol modcolor {self.rep_index} {self.index} colorID {self.get_color_id()}\n")
            file.write(f"mol modstyle {self.rep_index} {self.index} cpk 1.4\n")


mol = Molecules(path)

if args.file_name == fn_arg.default:
    for p in path.glob("marked_bead*.lammpstrj"):
        mol.create_new_marked(p.name)

kleisins = parameters["kleisin_ids"]
kleisin_rng = (min(kleisins), max(kleisins))
mol.create_new_dna(args.file_name, parameters["dna_indices_list"], [kleisin_rng])
mol.add_piece(kleisin_rng)

cmd = ["vmd", "-e", f"{mol.path.absolute()}"]
subprocess.run(cmd, cwd=path, check=True)



# Copyright (c) 2024 Lucas Dooms

from dataclasses import dataclass
from typing import List

import numpy as np

from smc_lammps.generate.generator import (
    BAI,
    AtomGroup,
    AtomType,
    BAI_Kind,
    BAI_Type,
    Generator,
    MoleculeId,
    PairWise,
)
from smc_lammps.generate.structures.smc.smc_creator import SMC_Pos


@dataclass
class SMC:
    use_rigid_hinge: bool

    pos: SMC_Pos

    t_arms_heads_kleisin: AtomType
    t_hinge: AtomType
    t_atp: AtomType
    t_upper_site: AtomType
    t_middle_site: AtomType
    t_lower_site: AtomType
    t_ref_site: AtomType

    # bonds
    k_bond: float
    k_hinge: float
    max_bond_length: float

    # angles
    # Bending of elbows (kinkable arms, hence soft)
    k_elbow: float
    # Arms opening angle wrt ATP bridge (should be stiff)
    k_arm: float

    # impropers
    # Fixes site orientation (prevents free rotation, should be stiff)
    k_align_site: float
    # Folding stiffness of lower compartment (should be stiff)
    k_fold: float
    # Makes folding asymmetric (should be stiff)
    k_asymmetry: float

    # other
    bridge_width: float
    arm_length: float
    hinge_radius: float
    arms_angle_ATP: float
    folding_angle_ATP: float
    folding_angle_APO: float

    def _set_molecule_ids(self, use_rigid_hinge: bool) -> None:
        # Molecule for each rigid body
        self.mol_arm_dl = MoleculeId.get_next()
        self.mol_arm_ul = MoleculeId.get_next()
        self.mol_arm_ur = MoleculeId.get_next()
        self.mol_arm_dr = MoleculeId.get_next()
        self.mol_heads_kleisin = MoleculeId.get_next()
        self.mol_ATP = MoleculeId.get_next()

        self.mol_hinge_l = MoleculeId.get_next()
        if use_rigid_hinge:
            self.mol_hinge_r = self.mol_hinge_l
        else:
            self.mol_hinge_r = MoleculeId.get_next()

        self.mol_middle_site = self.mol_ATP
        self.mol_lower_site = self.mol_heads_kleisin

    def get_molecule_ids(self) -> List[int]:
        return [
            self.mol_arm_dl,
            self.mol_arm_ul,
            self.mol_arm_ur,
            self.mol_arm_dr,
            self.mol_heads_kleisin,
            self.mol_ATP,
            self.mol_hinge_l,
            self.mol_hinge_r,
            self.mol_middle_site,
            self.mol_lower_site,
        ]

    def _set_angles(self) -> None:
        self.align_arms = BAI_Type(
            BAI_Kind.ANGLE, "harmonic", f"{self.k_elbow} {180.0}\n"
        )
        arms_bridge_angle = np.rad2deg(
            np.arccos(self.bridge_width / self.arm_length / 4.0)
        )
        self.arms_bridge = BAI_Type(
            BAI_Kind.ANGLE, "harmonic", f"{self.k_arm} {arms_bridge_angle}\n"
        )
        self.hinge_arms = BAI_Type(BAI_Kind.ANGLE, "harmonic", f"{self.k_arm} {90.0}\n")

        self.arms_close = Generator.DynamicCoeffs(
            BAI_Kind.ANGLE,
            f"harmonic {self.k_arm} {np.rad2deg(np.arccos((self.bridge_width / 2.0 - self.hinge_radius) / self.arm_length))}\n",
            [self.arms_bridge],
        )
        self.arms_open = Generator.DynamicCoeffs(
            BAI_Kind.ANGLE,
            f"harmonic {self.k_arm} {self.arms_angle_ATP}\n",
            [self.arms_bridge],
        )

    def _set_impropers(self) -> None:
        self.imp_t1 = BAI_Type(
            BAI_Kind.IMPROPER, "harmonic", f"{self.k_align_site} {0.0}\n"
        )
        self.imp_t2 = BAI_Type(
            BAI_Kind.IMPROPER,
            "harmonic",
            f"{self.k_fold} {180.0 - self.folding_angle_APO}\n",
        )
        self.imp_t3 = BAI_Type(
            BAI_Kind.IMPROPER,
            "harmonic",
            f"{self.k_asymmetry} {abs(90.0 - self.folding_angle_APO)}\n",
        )
        self.imp_t4 = BAI_Type(
            BAI_Kind.IMPROPER, "harmonic", f"{self.k_align_site / 5.0} {90.0}\n"
        )

        self.kleisin_folds1 = Generator.DynamicCoeffs(
            BAI_Kind.IMPROPER,
            f"{self.k_fold} {180.0 - self.folding_angle_ATP}\n",
            [self.imp_t2],
        )
        self.kleisin_unfolds1 = Generator.DynamicCoeffs(
            BAI_Kind.IMPROPER,
            f"{self.k_fold} {180.0 - self.folding_angle_APO}\n",
            [self.imp_t2],
        )

        self.kleisin_folds2 = Generator.DynamicCoeffs(
            BAI_Kind.IMPROPER,
            f"{self.k_asymmetry} {abs(90.0 - self.folding_angle_ATP)}\n",
            [self.imp_t3],
        )
        self.kleisin_unfolds2 = Generator.DynamicCoeffs(
            BAI_Kind.IMPROPER,
            f"{self.k_asymmetry} {abs(90.0 - self.folding_angle_APO)}\n",
            [self.imp_t3],
        )

    def __post_init__(self) -> None:
        self._set_molecule_ids(self.use_rigid_hinge)
        self._set_angles()
        self._set_impropers()
        # create groups
        self.arm_dl_grp = AtomGroup(
            self.pos.r_arm_dl, self.t_arms_heads_kleisin, self.mol_arm_dl
        )
        self.arm_ul_grp = AtomGroup(
            self.pos.r_arm_ul, self.t_arms_heads_kleisin, self.mol_arm_ul
        )
        self.arm_ur_grp = AtomGroup(
            self.pos.r_arm_ur, self.t_arms_heads_kleisin, self.mol_arm_ur
        )
        self.arm_dr_grp = AtomGroup(
            self.pos.r_arm_dr, self.t_arms_heads_kleisin, self.mol_arm_dr
        )
        self.hk_grp = AtomGroup(
            self.pos.r_kleisin, self.t_arms_heads_kleisin, self.mol_heads_kleisin
        )

        self.atp_grp = AtomGroup(self.pos.r_ATP, self.t_atp, self.mol_ATP)

        self.hinge_l_grp = AtomGroup(
            self.pos.r_hinge[: len(self.pos.r_hinge) // 2],
            self.t_hinge,
            self.mol_hinge_l,
            charge=-0.1,
        )
        self.hinge_r_grp = AtomGroup(
            self.pos.r_hinge[len(self.pos.r_hinge) // 2 :],
            self.t_hinge,
            self.mol_hinge_r,
            charge=-0.1,
        )
        self.upper_site_grp = AtomGroup(
            self.pos.r_upper_site, self.t_upper_site, self.mol_hinge_l
        )

        # split M in three parts

        cut = 2
        self.middle_site_grp = AtomGroup(
            self.pos.r_middle_site[:cut], self.t_middle_site, self.mol_middle_site
        )
        self.middle_site_atp_grp = AtomGroup(
            self.pos.r_middle_site[cut:-1], self.t_atp, self.mol_middle_site
        )
        # ref site
        self.middle_site_ref_grp = AtomGroup(
            self.pos.r_middle_site[-1:], self.t_ref_site, self.mol_middle_site
        )

        # split B in two parts

        cut = 3
        self.lower_site_grp = AtomGroup(
            self.pos.r_lower_site[:cut], self.t_lower_site, self.mol_lower_site
        )
        self.lower_site_arm_grp = AtomGroup(
            self.pos.r_lower_site[cut:], self.t_arms_heads_kleisin, self.mol_lower_site
        )

        self.left_attach_hinge = len(self.hinge_l_grp.positions) // 2
        self.right_attach_hinge = len(self.hinge_r_grp.positions) // 2

    def get_groups(self) -> List[AtomGroup]:
        return [
            self.arm_dl_grp,
            self.arm_ul_grp,
            self.arm_ur_grp,
            self.arm_dr_grp,
            self.hk_grp,
            self.atp_grp,
            self.upper_site_grp,
            self.middle_site_grp,
            self.middle_site_atp_grp,
            self.middle_site_ref_grp,
            self.lower_site_grp,
            self.lower_site_arm_grp,
            self.hinge_l_grp,
            self.hinge_r_grp,
        ]

    def get_bonds(self, hinge_opening: float | None = None) -> List[BAI]:
        # Every joint is kept in place through bonds
        attach = BAI_Type(
            BAI_Kind.BOND,
            "fene/expand",
            f"{self.k_bond} {self.max_bond_length} {0.0} {0.0} {0.0}\n",
        )

        bonds = [
            # attach arms together
            BAI(attach, (self.arm_dl_grp, -1), (self.arm_ul_grp, 0)),
            BAI(attach, (self.arm_ur_grp, -1), (self.arm_dr_grp, 0)),
            # attach hinge and arms
            BAI(
                attach,
                (self.arm_ul_grp, -1),
                (self.hinge_l_grp, self.left_attach_hinge),
            ),
            BAI(
                attach,
                (self.arm_ur_grp, 0),
                (self.hinge_r_grp, self.right_attach_hinge),
            ),
            # attach atp bridge to arms
            BAI(attach, (self.arm_dr_grp, -1), (self.atp_grp, -1)),
            BAI(attach, (self.atp_grp, 0), (self.arm_dl_grp, 0)),
            # attach bridge to hk
            BAI(attach, (self.atp_grp, -1), (self.hk_grp, 0)),
            BAI(attach, (self.hk_grp, -1), (self.atp_grp, 0)),
        ]

        if not self.use_rigid_hinge:
            assert hinge_opening is not None
            hinge_bond = BAI_Type(
                BAI_Kind.BOND, "harmonic", f"{self.k_hinge} {hinge_opening}\n"
            )
            bonds += [
                # connect Left and Right hinge pieces together
                BAI(hinge_bond, (self.hinge_l_grp, -1), (self.hinge_r_grp, 0)),
                BAI(hinge_bond, (self.hinge_l_grp, 0), (self.hinge_r_grp, -1)),
            ]
        return bonds

    def get_angles(self) -> List[BAI]:
        return [
            # keep left arms rigid (prevent too much bending)
            BAI(
                self.align_arms,
                (self.arm_dl_grp, 0),
                (self.arm_ul_grp, 0),
                (self.arm_ul_grp, -1),
            ),
            # same, but for right arms
            BAI(
                self.align_arms,
                (self.arm_ur_grp, 0),
                (self.arm_ur_grp, -1),
                (self.arm_dr_grp, -1),
            ),
            # keep hinge perpendicular to arms
            BAI(
                self.hinge_arms,
                (self.arm_ul_grp, -2),
                (self.arm_ul_grp, -1),
                (self.hinge_l_grp, self.left_attach_hinge + 1),
            ),
            BAI(
                self.hinge_arms,
                (self.arm_ul_grp, -2),
                (self.arm_ul_grp, -1),
                (self.hinge_l_grp, self.left_attach_hinge - 1),
            ),
            BAI(
                self.hinge_arms,
                (self.arm_ur_grp, 1),
                (self.arm_ur_grp, 0),
                (self.hinge_r_grp, self.right_attach_hinge - 1),
            ),
            BAI(
                self.hinge_arms,
                (self.arm_ur_grp, 1),
                (self.arm_ur_grp, 0),
                (self.hinge_r_grp, self.right_attach_hinge + 1),
            ),
            # prevent too much bending between lower arms and the bridge
            BAI(
                self.arms_bridge,
                (self.arm_dl_grp, -1),
                (self.arm_dl_grp, 0),
                (self.atp_grp, -1),
            ),
            BAI(
                self.arms_bridge,
                (self.arm_dr_grp, 0),
                (self.arm_dr_grp, -1),
                (self.atp_grp, 0),
            ),
        ]

    def get_impropers(self) -> List[BAI]:
        kleisin_center = len(self.pos.r_kleisin) // 2
        return [
            # Fix orientation of ATP/kleisin bridge
            # WARNING: siteM is split into groups, be careful with index
            BAI(
                self.imp_t1,
                (self.arm_dl_grp, -1),
                (self.arm_dl_grp, 0),
                (self.atp_grp, -1),
                (self.middle_site_grp, 1),
            ),
            BAI(
                self.imp_t1,
                (self.arm_dr_grp, 0),
                (self.arm_dr_grp, -1),
                (self.atp_grp, 0),
                (self.middle_site_grp, 1),
            ),
            BAI(
                self.imp_t2,
                (self.arm_dl_grp, -1),
                (self.arm_dl_grp, 0),
                (self.atp_grp, -1),
                (self.hk_grp, kleisin_center),
            ),
            BAI(
                self.imp_t2,
                (self.arm_dr_grp, 0),
                (self.arm_dr_grp, -1),
                (self.atp_grp, 0),
                (self.hk_grp, kleisin_center),
            ),
            # prevent kleisin ring from swaying too far relative to the bridge
            BAI(
                self.imp_t3,
                (self.middle_site_ref_grp, 0),
                (self.arm_dl_grp, 0),
                (self.arm_dr_grp, -1),
                (self.hk_grp, kleisin_center),
            ),
            # fix hinge to a plane
            BAI(
                self.imp_t1,
                (self.hinge_l_grp, 0),
                (self.hinge_l_grp, -1),
                (self.hinge_r_grp, 0),
                (self.hinge_r_grp, -1),
            ),
            BAI(
                self.imp_t1,
                (self.hinge_l_grp, 0),
                (self.hinge_l_grp, self.left_attach_hinge),
                (self.hinge_r_grp, 0),
                (self.hinge_r_grp, self.right_attach_hinge),
            ),
            # fix hinge perpendicular to arms plane
            BAI(
                self.imp_t4,
                (self.arm_ul_grp, -2),
                (self.arm_ul_grp, -1),
                (self.hinge_l_grp, self.left_attach_hinge - 1),
                (self.hinge_l_grp, self.left_attach_hinge + 1),
            ),
            BAI(
                self.imp_t4,
                (self.arm_ur_grp, 1),
                (self.arm_ur_grp, 0),
                (self.hinge_r_grp, self.right_attach_hinge + 1),
                (self.hinge_r_grp, self.right_attach_hinge - 1),
            ),
            # # keep hinge aligned with bridge axis
            # BAI(imp_t1, (self.hingeL_group, self.left_attach_hinge), (self.hingeR_group, self.right_attach_hinge), (self.atp_group, 0), (self.atp_group, -1)),
            BAI(
                self.imp_t4,
                (self.upper_site_grp, 0),
                (self.upper_site_grp, len(self.upper_site_grp.positions) // 2),
                (self.atp_grp, len(self.atp_grp.positions) // 2),
                (self.atp_grp, -1),
            ),
        ]

    def add_repel_interactions(
        self, pair_inter: PairWise, eps: float, sigma: float, r_cut: float
    ) -> None:
        # short-range repulsion
        sigma_short = sigma / 10.0
        r_cut_short = r_cut / 10.0

        # prevent hinges from overlapping
        pair_inter.add_interaction(
            self.t_hinge,
            self.t_hinge,
            eps,
            sigma_short,
            r_cut_short,
        )
        # prevent upper site from overlapping with arms
        pair_inter.add_interaction(
            self.t_arms_heads_kleisin,
            self.t_upper_site,
            eps,
            sigma_short,
            r_cut_short,
        )



# Copyright (c) 2024 Lucas Dooms

import math
from dataclasses import dataclass, fields
from typing import Any, List, Tuple

import numpy as np
from numpy.random import default_rng
from scipy.spatial.transform import Rotation

from smc_lammps.generate.generator import Nx3Array
from smc_lammps.generate.structures.structure_creator import (
    attach_chain,
    get_circle_segment_unit_radius,
    get_straight_segment,
)


@dataclass
class SMC_Pos:
    r_arm_dl: Nx3Array
    r_arm_ul: Nx3Array
    r_arm_ur: Nx3Array
    r_arm_dr: Nx3Array
    r_ATP: Nx3Array
    r_kleisin: Nx3Array
    r_upper_site: Nx3Array
    r_middle_site: Nx3Array
    r_lower_site: Nx3Array
    r_hinge: Nx3Array

    def iter(self) -> List[Any]:
        """Returns a list of all fields"""
        return [getattr(self, field.name) for field in fields(SMC_Pos)]

    def apply(self, func) -> None:
        """Update the object inplace by applying a function to every field"""
        for field in fields(self.__class__):
            setattr(self, field.name, func(getattr(self, field.name)))

    def map(self, func) -> List[Any]:
        """Apply a function to every field and return the resulting list"""
        return [func(x) for x in self.iter()]


@dataclass
class SMC_Creator:
    """Computes the positions of all SMC beads"""

    SMC_spacing: float

    # Horizontal distance between top binding sites (units bead spacing)
    upper_site_h: float
    # Vertical distance of top binding sites from hinge (units of bead spacing)
    upper_site_v: float
    # Horizontal distance between middle binding sites (units bead spacing)
    middle_site_h: float
    # Vertical distance of middle binding sites from bridge (units of bead spacing)
    middle_site_v: float
    # Horizontal distance between bottom binding sites (units bead spacing)
    lower_site_h: float
    # Distance of bottom binding sites from kleisin (units of bead spacing)
    lower_site_v: float

    arm_length: float
    bridge_width: float
    hinge_radius: float
    hinge_opening: float

    kleisin_radius: float

    folding_angle_APO: float

    small_noise: float = 1e-5

    def get_arms(
        self, seed: int = 8671288977726523465
    ) -> Tuple[Nx3Array, Nx3Array, Nx3Array, Nx3Array]:
        # Number of beads forming each arm segment (err on the high side)
        n_arm_segments = math.ceil(self.arm_length / (2 * self.SMC_spacing))

        # z and y lengths of each arm (2 aligned segments), for the initial triangular geometry
        z_arm = self.bridge_width / 2.0
        y_arm = math.sqrt(self.arm_length**2 - z_arm**2)

        # u: up, d: down
        # l: left, r: right
        direction_dl = [0, y_arm, z_arm]
        direction_ul = [0, y_arm, z_arm]
        direction_ur = [0, -y_arm, z_arm]
        direction_dr = [0, -y_arm, z_arm]

        factor = self.arm_length / 2.0 / n_arm_segments
        r_arm_dl = get_straight_segment(n_arm_segments, direction_dl) * factor
        r_arm_ul = get_straight_segment(n_arm_segments, direction_ul) * factor
        r_arm_ur = get_straight_segment(n_arm_segments, direction_ur) * factor
        r_arm_dr = get_straight_segment(n_arm_segments, direction_dr) * factor

        r_arm_dl, r_arm_ul, r_arm_ur, r_arm_dr = attach_chain(
            r_arm_dl, [[r_arm_ul, False], [r_arm_ur, False], [r_arm_dr, False]]
        )

        # move lower center to origin (where bridge will be placed)
        center = (r_arm_dl[0] + r_arm_dr[-1]) / 2.0

        r_arm_dl -= center
        r_arm_ul -= center
        r_arm_ur -= center
        r_arm_dr -= center

        # A bit of randomness, to avoid exact overlap (pressure is messed up in LAMMPS)
        rng_arms = default_rng(seed=seed)

        r_arm_dl += rng_arms.standard_normal() * self.small_noise
        r_arm_ul += rng_arms.standard_normal() * self.small_noise
        r_arm_ur += rng_arms.standard_normal() * self.small_noise
        r_arm_dr += rng_arms.standard_normal() * self.small_noise

        return r_arm_dl, r_arm_ul, r_arm_ur, r_arm_dr

    def get_bridge(self, seed: int = 4685150768879447999) -> Nx3Array:
        # Number of beads forming the ATP ring (err on the high side)
        n_ATP = math.ceil(self.bridge_width / self.SMC_spacing)

        # We want an odd number (necessary for angle/dihedral interactions)
        if n_ATP % 2 == 0:
            n_ATP += 1

        # Positions
        r_ATP = get_straight_segment(n_ATP, [0, 0, 1])

        # use the bridgeWidth as the absolute truth (SMC spacing may be slightly off)
        r_ATP *= self.bridge_width / n_ATP

        # move center to origin
        r_ATP -= (r_ATP[0] + r_ATP[-1]) / 2.0

        # A bit of randomness
        rng_atp = default_rng(seed=seed)
        r_ATP += rng_atp.standard_normal() * self.small_noise

        return r_ATP

    def get_heads_kleisin(self, seed: int = 8305832029550348799) -> Nx3Array:
        # Circle-arc radius
        # radius = (self.HKradius**2 + (self.bridgeWidth / 2.0)**2) / (2.0 * self.HKradius)
        bridge_radius = self.bridge_width / 2.0
        radius = self.kleisin_radius
        if radius < bridge_radius:
            raise ValueError(
                f"The kleisin radius ({radius}) is too small (< {bridge_radius}) based on the bridgeWidth {self.bridge_width}"
            )

        # from the y-axis
        starting_angle = math.asin(bridge_radius / radius)
        # Opening angle of circular arc (away from the bridge = 2*pi - angle towards the bridge)
        phi0 = 2.0 * math.pi - 2.0 * starting_angle

        # Number of beads forming the heads/kleisin complex (err on the high side)
        n_kleisin = math.ceil(phi0 * radius / self.SMC_spacing)

        # We want an odd number (necessary for angle/dihedral interactions)
        if n_kleisin % 2 == 0:
            n_kleisin += 1

        ending_angle = 2.0 * math.pi - starting_angle
        # add pi/2, since circle_segment calculation starts from x-axis
        starting_angle += math.pi / 2.0
        ending_angle += math.pi / 2.0

        r_kleisin = get_circle_segment_unit_radius(
            n_kleisin,
            end_inclusive=True,
            theta_start=starting_angle,
            theta_end=ending_angle,
            normal_direction=[1, 0, 0],
        )

        r_kleisin *= radius
        # move the bridge-gap to the origin
        r_kleisin[:, 1] -= r_kleisin[0][1]

        # A bit of randomness
        rng_rhk = default_rng(seed=seed)
        r_kleisin += rng_rhk.standard_normal() * self.small_noise

        return r_kleisin

    @staticmethod
    def shielded_site_template(
        n_inner_beads: int,
        n_outer_beads_per_inner_bead: int,
        inner_spacing: float,
        outer_spacing: float,
    ) -> Nx3Array:
        """create a line of beads surrounded by a protective shell/shield"""
        axis = np.array([1, 0, 0])
        # Inner/Attractive beads
        inner_beads = (
            get_straight_segment(n_inner_beads, direction=axis) * inner_spacing
        )
        # put center at the origin
        inner_beads -= (inner_beads[0] + inner_beads[-1]) / 2.0

        # Repulsive/Outer beads, forming a surrounding shell
        shells = []
        for inner_bead in inner_beads:
            shells.append(
                get_circle_segment_unit_radius(
                    n_outer_beads_per_inner_bead,
                    end_inclusive=True,
                    theta_start=0,
                    theta_end=np.pi,
                    normal_direction=axis,
                )
                * outer_spacing
            )
            # place center of shell at inner bead
            shells[-1] += inner_bead

        # Horizontal shield at two ends
        end_first = (inner_beads[0] - inner_spacing * axis).reshape(1, 3)
        end_last = (inner_beads[-1] + inner_spacing * axis).reshape(1, 3)

        return np.concatenate([inner_beads, *shells, end_first, end_last])

    @staticmethod
    def transpose_rotate_transpose(rotation, *arrays: Nx3Array) -> Tuple[Nx3Array]:
        return tuple(rotation.dot(arr.transpose()).transpose() for arr in arrays)

    def get_interaction_sites(
        self, lower_site_points_down: bool, seed: int = 8343859591397577529
    ) -> Tuple[Nx3Array, Nx3Array, Nx3Array]:
        # U = upper  interaction site
        # M = middle interaction site
        # D = lower  interaction site

        # UPPER SITE

        r_upper_site = get_straight_segment(3)
        r_upper_site -= r_upper_site[1]

        rotate_around_x_axis = Rotation.from_rotvec(
            math.pi * np.array([1.0, 0.0, 0.0])
        ).as_matrix()

        # MIDDLE SITE
        r_middle_site = self.shielded_site_template(1, 4, self.middle_site_h, 1)
        (r_middle_site,) = self.transpose_rotate_transpose(
            rotate_around_x_axis, r_middle_site
        )

        # take last bead and use it as an extra inner bead
        r_middle_site = np.concatenate(
            [r_middle_site[:1], r_middle_site[-1:], r_middle_site[1:-1]]
        )
        # move, so that this bead is at the origin
        r_middle_site -= r_middle_site[1]

        # LOWER SITE
        r_lower_site = self.shielded_site_template(3, 4, self.lower_site_h, 1)
        if not lower_site_points_down:
            (r_lower_site,) = self.transpose_rotate_transpose(
                rotate_around_x_axis, r_lower_site
            )

        # Add randomness
        rng_sites = default_rng(seed=seed)
        r_upper_site += rng_sites.standard_normal() * self.small_noise
        r_middle_site += rng_sites.standard_normal() * self.small_noise
        r_lower_site += rng_sites.standard_normal() * self.small_noise

        return r_upper_site, r_middle_site, r_lower_site

    def get_hinge(self) -> Nx3Array:
        radius = self.hinge_radius

        spacing = self.SMC_spacing * 0.8

        n_ring = math.ceil(2 * np.pi * radius / spacing)
        # should be multiple of 2 but not of 4
        if n_ring % 2 == 1:
            n_ring -= 1
        if n_ring % 4 == 0:
            n_ring += 2

        r_hinge = get_circle_segment_unit_radius(
            n_ring, end_inclusive=False, normal_direction=(0, 1, 0)
        )

        # rotate slightly
        angle = np.linalg.norm(r_hinge[1] - r_hinge[0]) / 2.0
        rotation = Rotation.from_rotvec(angle * np.array([0.0, 1.0, 0.0])).as_matrix()

        r_hinge *= radius

        (r_hinge,) = self.transpose_rotate_transpose(rotation, r_hinge)

        # separate pieces
        half = len(r_hinge) // 2
        # overlap
        r_hinge[:half, 2] -= r_hinge[0, 2]
        r_hinge[half:, 2] -= r_hinge[-1, 2]

        r_hinge[:half, 2] -= self.hinge_opening / 2.0
        r_hinge[half:, 2] += self.hinge_opening / 2.0

        return r_hinge

    def get_smc(self, lower_site_points_down: bool, extra_rotation=None) -> SMC_Pos:
        r_arm_dl, r_arm_ul, r_arm_ur, r_arm_dr = self.get_arms()
        r_ATP = self.get_bridge()
        r_kleisin = self.get_heads_kleisin()
        r_upper_site, r_middle_site, r_lower_site = self.get_interaction_sites(
            lower_site_points_down
        )
        r_hinge = self.get_hinge()

        # Inert bead, used for breaking folding symmetry
        r_middle_site = np.concatenate(
            [r_middle_site, np.array([1.0, -1.0, 0.0]).reshape(1, 3)]
        )
        r_middle_site[:, 1] += self.middle_site_v
        if lower_site_points_down:
            r_lower_site[:, 1] -= self.lower_site_v
        else:
            r_lower_site[:, 1] += self.lower_site_v

        # scale properly
        r_upper_site *= self.SMC_spacing
        r_middle_site *= self.SMC_spacing
        r_lower_site *= self.SMC_spacing

        # place hinge at center of top
        r_hinge += r_arm_ur[0]
        # place bead slightly below
        r_upper_site += r_arm_ur[0]
        r_upper_site[:, 1] -= self.SMC_spacing

        # rotate upper arms away to attach to hinge properly
        left_attach_hinge = len(r_hinge) // 4
        rot = Rotation.align_vectors(
            r_arm_ul[-1] - r_arm_ul[0], r_hinge[left_attach_hinge] - r_arm_ul[0]
        )[0]
        r_arm_ur = (
            self.transpose_rotate_transpose(rot.as_matrix(), r_arm_ur - r_arm_ur[-1])[0]
            + r_arm_ur[-1]
        )
        r_arm_ul = (
            self.transpose_rotate_transpose(
                rot.inv().as_matrix(), r_arm_ul - r_arm_ul[0]
            )[0]
            + r_arm_ul[0]
        )

        # move into the correct location
        r_middle_site += r_ATP[len(r_ATP) // 2]
        r_lower_site += r_kleisin[len(r_kleisin) // 2]

        ############################# Fold upper compartment ############################

        # Rotation matrix (clockwise about z axis)
        rotation = Rotation.from_rotvec(
            -math.radians(self.folding_angle_APO) * np.array([0.0, 0.0, 1.0])
        ).as_matrix()

        # Rotate upper segments only
        r_arm_dl, r_arm_ul, r_arm_ur, r_arm_dr, r_upper_site, r_middle_site, r_hinge = (
            self.transpose_rotate_transpose(
                rotation,
                r_arm_dl,
                r_arm_ul,
                r_arm_ur,
                r_arm_dr,
                r_upper_site,
                r_middle_site,
                r_hinge,
            )
        )

        self.generated_positions = SMC_Pos(
            r_arm_dl=r_arm_dl,
            r_arm_ul=r_arm_ul,
            r_arm_ur=r_arm_ur,
            r_arm_dr=r_arm_dr,
            r_ATP=r_ATP,
            r_kleisin=r_kleisin,
            r_upper_site=r_upper_site,
            r_middle_site=r_middle_site,
            r_lower_site=r_lower_site,
            r_hinge=r_hinge,
        )

        # apply extra rotation to entire SMC
        if extra_rotation is not None:
            rotation = Rotation.from_rotvec(extra_rotation).as_matrix()
            self.generated_positions.apply(
                lambda pos: self.transpose_rotate_transpose(rotation, pos)
            )

        return self.generated_positions

    def get_mass_per_atom(self, total_mass: float) -> float:
        return total_mass / sum(self.generated_positions.map(len))



# Copyright (c) 2024 Lucas Dooms

"""
generator.py
------------
How to use:
    - create a Generator instance
    - create AtomType instances for each type you need
    - create BAI_Type for Bond/Angle/Improper types
    - define AtomGroup with positions and type
    - define BAI for interactions between specific atoms
    - define PairWise, and use add_interaction to define global interactions
    - append all defined AtomGroup, BAI, and PairWise instances to the lists in Generator
    - call Generator.write(file) to create the final datafile
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Set, Tuple

import numpy as np
import numpy.typing as npt

from smc_lammps.console import warn


class AtomType:
    __index = 0

    def __init__(self, mass: float = 1.0) -> None:
        self._index = None
        self.mass = mass

    @classmethod
    def _get_next(cls) -> int:
        cls.__index += 1
        return cls.__index

    @property
    def index(self) -> int:
        if self._index is None:
            self._index = self._get_next()
        return self._index


class MoleculeId:
    __index = 0

    @classmethod
    def get_next(cls) -> int:
        cls.__index += 1
        return cls.__index


class BAI_Kind(Enum):
    BOND = 1
    ANGLE = 2
    IMPROPER = 3

    @classmethod
    def length_lookup(cls, kind: BAI_Kind) -> int:
        """Returns the number of arguments (atom ids) needed to define a BAI interaction."""
        values = {BAI_Kind.BOND: 2, BAI_Kind.ANGLE: 3, BAI_Kind.IMPROPER: 4}
        return values[kind]


class BAI_Type:
    indices = {kind: 1 for kind in BAI_Kind}

    def __init__(self, kind: BAI_Kind, style: str, coefficients: str = "") -> None:
        """coefficients will be printed after the index in a datafile
        e.g. 3 harmonic 2 5
        (where 3 is the index and the rest is the coefficients string)"""
        self._index = None
        self.kind = kind
        self.style = style
        self.coefficients = coefficients

    @property
    def index(self) -> int:
        if self._index is None:
            self._index = self.indices[self.kind]
            self.indices[self.kind] += 1
        return self._index

    def get_string(self, omit_style=False) -> str:
        style = f" {self.style}" if not omit_style else ""
        return f"{self.index}{style} {self.coefficients}"


Nx3Array = npt.NDArray[np.float32]
"""An (N, 3) array of positions"""


class AtomGroup:
    """
    positions: numpy array with shape (n, 3) of 3d positions of (n) atoms [[x, y, z], ...]
    atom_type: type of the atoms
    molecule_index (int): the molecule index which is needed for atom_style molecule
    polymer_bond_type: if None -> no bonds, otherwise all atoms will form bonds as a polymer
    """

    def __init__(
        self,
        positions: Nx3Array,
        atom_type: AtomType,
        molecule_index: int,
        polymer_bond_type: BAI_Type | None = None,
        polymer_angle_type: BAI_Type | None = None,
        charge: float = 0.0,
    ) -> None:
        self.positions = positions
        self.type = atom_type
        self.molecule_index = molecule_index
        if polymer_bond_type is not None:
            assert polymer_bond_type.kind == BAI_Kind.BOND
        self.polymer_bond_type = polymer_bond_type
        if polymer_angle_type is not None:
            assert polymer_angle_type.kind == BAI_Kind.ANGLE
        self.polymer_angle_type = polymer_angle_type
        self.charge = charge

    @property
    def n(self) -> int:
        return len(self.positions)


AtomIdentifier = Tuple[AtomGroup, int]
"A unique identifier of an atom in a group"


class BAI:
    """
    Represents a Bond/Angle/Improper interaction between a certain number of atoms
    """

    def __init__(self, type_: BAI_Type, *atoms: AtomIdentifier) -> None:
        """Length of atoms should be 2 for Bond, 3 for Angle, 4 for Improper"""
        self.type = type_
        self.atoms: List[AtomIdentifier] = list(atoms)


class PairWise:
    """
    Represents pair interactions between all atoms of two atoms ids.

    header: string defining the interaction style, e.g. "PairIJ Coeffs # hybrid"
    template: a string with empty formatters `{}` for the interaction parameters, e.g. "lj/cut {} {} {}"
    default: a list of default parameters (used to fill out all interactions, since LAMMPS requires them
                                           to all be explicitly defined)
    """

    def __init__(self, header: str, template: str, default: List[Any] | None) -> None:
        """if default is None -> don't insert missing interactions"""
        self.header = header
        self.template = template
        self.default = default
        self.pairs: List[Tuple[AtomType, AtomType, List[Any]]] = []

    def add_interaction(
        self, atom_type1: AtomType, atom_type2: AtomType, *args: Any
    ) -> PairWise:
        """Add an iteraction. Will sort the indices automatically."""
        ordered = sorted([atom_type1, atom_type2], key=lambda at: at.index)
        self.pairs.append((ordered[0], ordered[1], list(args)))

        return self

    def write(self, file, atom_types: List[AtomType]) -> None:
        """Write the Pair Coeffs header and all pair interactions to a file.
        e.g.
        PairIJ Coeffs
        1 8 lj/cut 0.0 0.0 0.0
        """
        file.write(self.header)
        for atom_type1, atom_type2, text in self.get_all_interactions(atom_types):
            file.write(f"{atom_type1.index} {atom_type2.index} " + text)
        file.write("\n")

    def get_all_interaction_pairs(
        self, all_atom_types: List[AtomType]
    ) -> List[Tuple[AtomType, AtomType]]:
        """Returns all possible interactions, whether they are set by the user or not."""
        present_atom_types = set()
        for pair in self.pairs:
            present_atom_types.add(pair[0])
            present_atom_types.add(pair[1])

        all_inters: List[Tuple[AtomType, AtomType]] = []
        all_atom_types = sorted(all_atom_types, key=lambda atom_type: atom_type.index)
        for i in range(len(all_atom_types)):
            for j in range(i, len(all_atom_types)):
                all_inters.append((all_atom_types[i], all_atom_types[j]))

        return all_inters

    def get_all_interactions(
        self, all_atom_types: List[AtomType]
    ) -> List[Tuple[AtomType, AtomType, str]]:
        """Returns actual interactions to define,
        applying the default where no interaction was specified by the user."""
        all_inters = self.get_all_interaction_pairs(all_atom_types)

        def pair_in_inter(
            interaction: Tuple[AtomType, AtomType],
        ) -> Tuple[AtomType, AtomType, List[Any]] | None:
            for pair in self.pairs:
                if interaction[0] == pair[0] and interaction[1] == pair[1]:
                    return pair
                if interaction[1] == pair[0] and interaction[0] == pair[1]:
                    return pair

            return None

        final_pairs: List[Tuple[AtomType, AtomType, str]] = []

        for inter in all_inters:
            pair = pair_in_inter(inter)
            if pair is None:
                if self.default is not None:
                    final_pairs.append(
                        (inter[0], inter[1], self.template.format(*self.default))
                    )
            else:
                final_pairs.append((pair[0], pair[1], self.template.format(*pair[2])))

        return final_pairs


def write_if_non_zero(file, fmt_string: str, amount: int):
    if amount != 0:
        file.write(fmt_string.format(amount))


class Generator:
    def __init__(self) -> None:
        self.atom_groups: List[AtomGroup] = []
        self.bais: List[BAI] = []
        self.atom_group_map: List[int] = []
        self.pair_interactions: List[PairWise] = []
        self.box_width = None
        self.molecule_override: Dict[AtomIdentifier, int] = {}
        self.charge_override: Dict[AtomIdentifier, float] = {}
        self.use_charges = False
        self.hybrid_styles = {
            BAI_Kind.BOND: "hybrid",
            BAI_Kind.ANGLE: "hybrid",
            BAI_Kind.IMPROPER: "hybrid",
        }
        self.random_shift = lambda: np.array([0.0, 0.0, 0.0])
        """Function that returns a random shift vector.
        This is useful to avoid exact overlap, which causes LAMMPS to crash during kspace calculations (e.g. with pair_style coul)."""

    def set_system_size(self, box_width: float) -> None:
        """Set the box size of the simulation."""
        self.box_width = box_width

    def get_total_atoms(self) -> int:
        """Get the total number of atoms across all groups."""
        return sum(map(lambda atom_group: atom_group.n, self.atom_groups))

    def write_header(self, file) -> None:
        """Write the top header to a file.
        Note: LAMMPS always ignores the first line of a data file,
        which should be this header."""
        file.write("# LAMMPS data file\n")

    def get_all_atom_types(self) -> List[AtomType]:
        """Get a list of all atom types across all atom groups.
        The returned list is sorted based on the AtomType index."""
        atom_types: Set[AtomType] = set()
        for atom_group in self.atom_groups:
            atom_types.add(atom_group.type)
        return sorted(atom_types, key=lambda atom_type: atom_type.index)

    def get_all_types(self, kind: BAI_Kind) -> List[BAI_Type]:
        """Get a list of all Bond/Angle/Improper types across all atom groups.
        The returned list is sorted based on the BAI_Type index."""
        bai_types: Set[BAI_Type] = set()
        for bai in filter(lambda bai: bai.type.kind == kind, self.bais):
            bai_types.add(bai.type)
        for atom_group in self.atom_groups:
            if kind == BAI_Kind.BOND and atom_group.polymer_bond_type is not None:
                bai_types.add(atom_group.polymer_bond_type)
            if kind == BAI_Kind.ANGLE and atom_group.polymer_angle_type is not None:
                bai_types.add(atom_group.polymer_angle_type)
        return sorted(bai_types, key=lambda bai_type: bai_type.index)

    def get_bai_dict_by_type(self) -> Dict[BAI_Kind, List[BAI]]:
        """Get a dictionary mapping the Bond/Angle/Improper kind
        to a list of all BAIs which have that type."""
        bai_by_kind: Dict[BAI_Kind, List[BAI]] = {kind: [] for kind in BAI_Kind}
        for bai in self.bais:
            bai_by_kind[bai.type.kind].append(bai)
        return bai_by_kind

    def get_amounts(self) -> Tuple[int, int, int]:
        """Return the total amount of bonds, angles, and impropers."""
        length_lookup = {
            key: len(value) for (key, value) in self.get_bai_dict_by_type().items()
        }

        total_bonds = length_lookup[BAI_Kind.BOND]
        total_angles = length_lookup[BAI_Kind.ANGLE]
        total_impropers = length_lookup[BAI_Kind.IMPROPER]

        for atom_group in self.atom_groups:
            if atom_group.polymer_bond_type is not None:
                total_bonds += len(atom_group.positions) - 1
            if atom_group.polymer_angle_type is not None:
                total_angles += len(atom_group.positions) - 2

        return total_bonds, total_angles, total_impropers

    def write_amounts(self, file) -> None:
        """Write the amount of atoms, bonds, angles, and impropers to a file."""
        file.write(f"{self.get_total_atoms()} atoms\n")

        total_bonds, total_angles, total_impropers = self.get_amounts()

        write_if_non_zero(file, "{} bonds\n", total_bonds)
        write_if_non_zero(file, "{} angles\n", total_angles)
        write_if_non_zero(file, "{} impropers\n", total_impropers)

        file.write("\n")

    def write_types(self, file) -> None:
        """Write the amount of atom types, bond types, angle types, and improper types to a file."""

        write_if_non_zero(file, "{} atom types\n", len(self.get_all_atom_types()))
        write_if_non_zero(
            file, "{} bond types\n", len(self.get_all_types(BAI_Kind.BOND))
        )
        write_if_non_zero(
            file, "{} angle types\n", len(self.get_all_types(BAI_Kind.ANGLE))
        )
        write_if_non_zero(
            file, "{} improper types\n", len(self.get_all_types(BAI_Kind.IMPROPER))
        )

        file.write("\n")

    def write_system_size(self, file) -> None:
        """Write the system size to a file."""
        file.write("# System size\n")

        if self.box_width is None:
            raise TypeError("box_width was not set")
        half_width = self.box_width / 2.0
        lohi = (-half_width, half_width)
        file.write(f"{lohi[0]} {lohi[1]} xlo xhi\n")
        file.write(f"{lohi[0]} {lohi[1]} ylo yhi\n")
        file.write(f"{lohi[0]} {lohi[1]} zlo zhi\n")

        file.write("\n")

    def write_masses(self, file) -> None:
        """Write the masses of all atom types to a file."""
        file.write("Masses\n\n")
        for atom_type in self.get_all_atom_types():
            file.write(f"{atom_type.index} {atom_type.mass}\n")

        file.write("\n")

    def get_all_BAI_styles(self) -> Dict[BAI_Kind, List[str]]:
        """Return a list of unique styles for each BAI kind."""

        def get_unique_styles(bai_types: List[BAI_Type]) -> List[str]:
            return list(set(t.style for t in bai_types))

        return {k: get_unique_styles(self.get_all_types(k)) for k in BAI_Kind}

    @staticmethod
    def get_BAI_style_command_name(kind: BAI_Kind) -> str:
        lookup = {
            BAI_Kind.BOND: "bond_style",
            BAI_Kind.ANGLE: "angle_style",
            BAI_Kind.IMPROPER: "improper_style",
        }
        return lookup[kind]

    def get_BAI_styles_command(self) -> str:
        """Return the command to define the BAI styles in a LAMMPS script."""
        all_styles = self.get_all_BAI_styles()

        def extract_command(k: BAI_Kind) -> str:
            styles = all_styles[k]
            if len(styles) == 1:
                return styles[0]
            styles_string = " ".join(styles)
            return f"{self.hybrid_styles[k]} {styles_string}"

        strings = [
            f"{self.get_BAI_style_command_name(k)} {extract_command(k)}\n"
            for k in BAI_Kind
        ]
        return "".join(strings)

    def get_hybrid_or_single_style(self) -> Dict[BAI_Kind, str]:
        """Return the style needed for the BAI Coeffs header."""

        def extract_style(k: BAI_Kind, styles: List[str]) -> str:
            if len(styles) == 1:
                return styles[0]
            return self.hybrid_styles[k]

        all_styles = self.get_all_BAI_styles()
        return {k: extract_style(k, v) for k, v in all_styles.items()}

    def get_BAI_coeffs_header(self, kind: BAI_Kind) -> str:
        """Get the header string corresponding to a Bond/Angle/Improper kind."""
        lookup = {
            BAI_Kind.BOND: "Bond Coeffs # {}\n\n",
            BAI_Kind.ANGLE: "Angle Coeffs # {}\n\n",
            BAI_Kind.IMPROPER: "Improper Coeffs # {}\n\n",
        }
        style_strings = self.get_hybrid_or_single_style()
        lookup = {k: v.format(style_strings[k]) for k, v in lookup.items()}
        return lookup[kind]

    def write_BAI_coeffs(self, file) -> None:
        """Write the Bond/Angle/Improper coefficients for each BAI kind to a file."""
        total_bonds, total_angles, total_impropers = self.get_amounts()
        lookup = {
            BAI_Kind.BOND: total_bonds,
            BAI_Kind.ANGLE: total_angles,
            BAI_Kind.IMPROPER: total_impropers,
        }
        for kind in BAI_Kind:
            if lookup[kind] == 0:
                # do not write anything if there are no BAIs of this kind
                continue

            file.write(self.get_BAI_coeffs_header(kind))
            all_types = self.get_all_types(kind)
            for bai_type in all_types:
                if not bai_type.coefficients:
                    continue
                omit_style = len(set(t.style for t in all_types)) == 1
                file.write(bai_type.get_string(omit_style))
            file.write("\n")

    def write_pair_interactions(self, file) -> None:
        """Write the Pair Coeffs header(s) and corresponding pair interactions to a file."""
        all_atom_types = self.get_all_atom_types()
        for pair in self.pair_interactions:
            pair.write(file, all_atom_types)

    def get_atom_index(self, atom_id: AtomIdentifier) -> int:
        """Get the absolute LAMMPS index for an atom.
        You must call write_atoms before using this method."""
        if not self.atom_group_map:
            raise AttributeError("write_atoms must be called first")

        index = self.atom_groups.index(atom_id[0])
        if atom_id[1] < 0:
            atom_group_length = len(atom_id[0].positions)
            atom_id = (atom_id[0], atom_id[1] + atom_group_length)
        return self.atom_group_map[index] + atom_id[1]

    def _set_up_atom_group_map(self) -> None:
        index_offset = 1
        for atom_group in self.atom_groups:
            self.atom_group_map.append(index_offset)
            index_offset += len(atom_group.positions)

    def get_atom_style(self) -> str:
        """Return the atom_style for LAMMPS (based on use_charges)."""
        if self.use_charges:
            return "full"
        else:
            return "molecular"

    def get_atoms_header(self) -> str:
        return f"Atoms # {self.get_atom_style()}\n\n"

    def get_atom_style_command(self) -> str:
        return f"atom_style {self.get_atom_style()}\n"

    def check_charges(self) -> None:
        """Checks for discrepancies between the use_charges flag and the actual atom charges."""
        nonzero_charges = [grp.charge != 0.0 for grp in self.atom_groups]

        if self.use_charges:
            if not any(nonzero_charges):
                warn(
                    "Charges are enabled, but all atoms have zero charge, this may affect performance!\n"
                    "Set use_charges=False to disable charges."
                )
        else:
            if any(nonzero_charges):
                warn(
                    "Charges are disabled, but some atoms have nonzero charge!\n"
                    "Set use_charges=True to enable charges."
                )

    def write_atoms(self, file) -> None:
        """Write the Atoms header and all atom positions to a file."""
        self.check_charges()
        file.write(self.get_atoms_header())

        self._set_up_atom_group_map()
        molecule_override_ids = {
            self.get_atom_index(atom_id): mol_id
            for atom_id, mol_id in self.molecule_override.items()
        }
        charge_override_values = {
            self.get_atom_index(atom_id): charge
            for atom_id, charge in self.charge_override.items()
        }

        for atom_group in self.atom_groups:
            for j, position in enumerate(atom_group.positions):
                atom_id = self.get_atom_index((atom_group, j))
                try:
                    mol_id = molecule_override_ids[atom_id]
                except KeyError:
                    mol_id = atom_group.molecule_index

                ids = f"{atom_id} {mol_id} {atom_group.type.index}"

                if self.use_charges:
                    try:
                        charge = charge_override_values[atom_id]
                    except KeyError:
                        charge = atom_group.charge
                    ids += f" {charge}"

                position += self.random_shift()
                file.write(ids + f" {position[0]} {position[1]} {position[2]}\n")

        file.write("\n")

    @staticmethod
    def get_BAI_header(kind: BAI_Kind) -> str:
        lookup = {
            BAI_Kind.BOND: "Bonds\n\n",
            BAI_Kind.ANGLE: "Angles\n\n",
            BAI_Kind.IMPROPER: "Impropers\n\n",
        }
        return lookup[kind]

    def write_bai(self, file) -> None:
        """Write the Bond/Angle/Improper headers and all corresponding BAI interactions to a file."""
        total_bonds, total_angles, total_impropers = self.get_amounts()
        lookup = {
            BAI_Kind.BOND: total_bonds,
            BAI_Kind.ANGLE: total_angles,
            BAI_Kind.IMPROPER: total_impropers,
        }

        for kind in BAI_Kind:
            if lookup[kind] == 0:
                # do not write anything if there are no BAIs of this kind
                continue

            file.write(self.get_BAI_header(kind))

            global_index = 1

            for atom_group in self.atom_groups:
                if kind == BAI_Kind.BOND and atom_group.polymer_bond_type is not None:
                    for j in range(len(atom_group.positions) - 1):
                        file.write(
                            f"{global_index} {atom_group.polymer_bond_type.index} "
                            f"{self.get_atom_index((atom_group, j))} "
                            f"{self.get_atom_index((atom_group, j + 1))}"
                            "\n"
                        )
                        global_index += 1

                if kind == BAI_Kind.ANGLE and atom_group.polymer_angle_type is not None:
                    for j in range(len(atom_group.positions) - 2):
                        file.write(
                            f"{global_index} {atom_group.polymer_angle_type.index} "
                            f"{self.get_atom_index((atom_group, j))} "
                            f"{self.get_atom_index((atom_group, j + 1))} "
                            f"{self.get_atom_index((atom_group, j + 2))} "
                            "\n"
                        )
                        global_index += 1

            length = BAI_Kind.length_lookup(kind)
            for bai in [bai for bai in self.bais if bai.type.kind == kind]:
                file.write(f"{global_index} {bai.type.index} ")
                formatter = ("{} " * length)[:-1] + "\n"
                file.write(
                    formatter.format(
                        *(self.get_atom_index(bai.atoms[i]) for i in range(length))
                    )
                )
                global_index += 1

            file.write("\n")

    def write_coeffs(self, file) -> None:
        """Write the coefficient information to a file.
        Useful when restarting a simulation."""
        self.write_header(file)
        self.write_types(file)
        file.write("\n")
        self.write_masses(file)
        self.write_BAI_coeffs(file)
        self.write_pair_interactions(file)

    def write_positions_and_bonds(self, file) -> None:
        """Write the positions and bonds to a file."""
        self.write_header(file)
        self.write_amounts(file)
        self.write_types(file)
        self.write_system_size(file)
        file.write("\n")
        self.write_atoms(file)
        self.write_bai(file)

    def write_full(self, file) -> None:
        """Write a full LAMMPS data file."""
        self.write_header(file)
        self.write_amounts(file)
        self.write_types(file)
        self.write_system_size(file)
        file.write("\n")
        self.write_masses(file)
        self.write_BAI_coeffs(file)
        self.write_pair_interactions(file)
        self.write_atoms(file)
        self.write_bai(file)

    @staticmethod
    def get_script_bai_command_name(pair_or_BAI: BAI_Kind | None) -> str:
        """Get the command name for a pair / BAI interaction.
        Used to redefine coefficients within a LAMMPS script."""
        name_dict = {
            None: "pair_coeff",
            BAI_Kind.BOND: "bond_coeff",
            BAI_Kind.ANGLE: "angle_coeff",
            BAI_Kind.IMPROPER: "improper_coeff",
        }
        return name_dict[pair_or_BAI]

    @dataclass
    class DynamicCoeffs:
        pair_or_BAI: None | BAI_Kind
        coeff_string: str
        args: List[Any]

    def write_script_bai_coeffs(self, file, coeffs: DynamicCoeffs) -> None:
        """Write a LAMMPS command for a pair / BAI interaction to a file."""
        cmd_name = self.get_script_bai_command_name(coeffs.pair_or_BAI)
        # parse args:
        # if pair type -> get atom indices (two AtomType instances)
        # else (bai) -> get BAI_Type index
        # in both cases -> assume arguments have .index field
        format_args = [str(arg.index) for arg in coeffs.args]
        formatted_string = " ".join(format_args) + " " + coeffs.coeff_string
        file.write(cmd_name + " " + formatted_string)


def test_simple_atoms():
    positions = np.zeros(shape=(100, 3), dtype=np.float32)
    gen = Generator()
    gen.atom_groups.append(AtomGroup(positions, AtomType(), 1))
    gen.set_system_size(10)
    with open("test.gen", "w", encoding="utf-8") as file:
        gen.write_full(file)


def test_simple_atoms_polymer():
    positions = np.zeros(shape=(100, 3), dtype=np.float32)
    gen = Generator()
    gen.atom_groups.append(
        AtomGroup(
            positions, AtomType(), 1, polymer_bond_type=BAI_Type(BAI_Kind.BOND, "fene")
        )
    )
    gen.set_system_size(10)
    with open("test.gen", "w", encoding="utf-8") as file:
        gen.write_full(file)


def test_with_bonds():
    positions = np.zeros(shape=(25, 3), dtype=np.float32)

    gen = Generator()
    gen.set_system_size(10)

    bt1 = BAI_Type(BAI_Kind.BOND, "harmonic")
    bt2 = BAI_Type(BAI_Kind.BOND, "harmonic")

    group1 = AtomGroup(positions, AtomType(), 1, polymer_bond_type=bt2)
    gen.atom_groups.append(group1)

    group2 = AtomGroup(np.copy(positions), AtomType(), 3)
    gen.atom_groups.append(group2)

    gen.bais.append(BAI(bt1, (group1, 1), (group2, 0)))

    gen.bais.append(BAI(BAI_Type(BAI_Kind.BOND, "cosine"), (group1, 5), (group1, 6)))

    gen.bais.append(BAI(bt1, (group1, 9), (group1, 16)))

    with open("test.gen", "w", encoding="utf-8") as file:
        gen.write_full(file)


def test_with_pairs():
    positions = np.zeros(shape=(25, 3), dtype=np.float32)

    gen = Generator()
    gen.set_system_size(10)

    at1 = AtomType()
    group1 = AtomGroup(positions, at1, 1)
    gen.atom_groups.append(group1)

    group2 = AtomGroup(np.copy(positions), AtomType(), 3)
    gen.atom_groups.append(group2)

    pairwise = PairWise("PairIJ Coeffs # hybrid\n", "lj/cut {} {} {}\n", [0, 0, 0])
    pairwise.add_interaction(at1, at1, 1, 2, 3)

    gen.pair_interactions.append(pairwise)

    with open("test.gen", "w", encoding="utf-8") as file:
        gen.write_full(file)


def all_tests():
    test_simple_atoms()
    test_simple_atoms_polymer()
    test_with_bonds()
    test_with_pairs()


def main():
    all_tests()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
import argparse
import subprocess
from functools import partial
from pathlib import Path
from re import compile as compile_regex
from typing import List

import argcomplete
from click import confirm

from smc_lammps.console import warn
from smc_lammps.generate.util import get_project_root

PYRUN = ["python", "-m"]


def parse() -> argparse.Namespace:
    # fmt: off
    parser = argparse.ArgumentParser(
        description='runs setup scripts, LAMMPS script, post-processing, and visualization',
        epilog='visit https://github.com/LucasDooms/SMC_LAMMPS for more info'
    )

    parser.add_argument('directory', help='the directory containing parameters for LAMMPS')

    generate_and_run = parser.add_argument_group(title='generate & run')
    generate_and_run.add_argument('-g', '--generate', action='store_true', help='run the python setup scripts before executing LAMMPS')
    generate_and_run.add_argument('-r', '--run', action='store_true', help='run the LAMMPS script')
    generate_and_run.add_argument('-c', '--continue', dest='continue_flag', action='store_true', help='continue from restart file and append to existing simulation')

    gar_mods = parser.add_argument_group(title='modifiers')
    gar_mods.add_argument('-s', '--seed', help='set the seed to be used by LAMMPS, this takes precedence over the seed in default_parameters.py and parameters.py (Note: currently only works with the --generate flag)')
    gar_mods.add_argument('-e', '--executable', help='name of the LAMMPS executable to use, default: \'lmp\'', default='lmp')
    gar_mods.add_argument('-f', '--force', action='store_true', help='don\'t prompt before overwriting existing files / continuing empty simulation')
    gar_mods.add_argument('-o', '--output', help='path to dump LAMMPS output to (prints to terminal by default)')
    gar_mods.add_argument('-sf', '--suffix', help='variant of LAMMPS styles to use, default: \'opt\' (see https://docs.lammps.org/Run_options.html#suffix)', default='opt')

    post_processing = parser.add_argument_group(title='post-processing')
    post_processing.add_argument('-p', '--post-process', action='store_true', help='run the post-processing scripts after running LAMMPS')
    pp_vis = post_processing.add_mutually_exclusive_group()
    pp_vis.add_argument('-v', '--visualize', action='store_true', help='open VMD after all scripts have finished')
    pp_vis.add_argument('-vd', '--visualize-datafile', action='store_true', help='shows the initial structure in VMD')
    pp_vis.add_argument('-vf', '--visualize-follow', action='store_true', help='same as --visualize, but follows the SMC')

    other = parser.add_argument_group(title='other options')
    other.add_argument('-n', '--ignore-errors', action='store_true', help='keep running even if the previous script exited with a non-zero error code')
    other.add_argument('-i', '-in', '--input', help='path to input file to give to LAMMPS')
    other.add_argument('--clean', action='store_true', help='remove all files except parameters.py from the directory')

    # fmt: on

    # shell autocompletion
    argcomplete.autocomplete(parser)

    return parser.parse_args()


def run_and_handle_error(process, ignore_errors: bool):
    completion: subprocess.CompletedProcess = process()
    if completion.returncode != 0:
        message = (
            f"\n\nprocess ended with error code {completion.returncode}\n{completion}\n"
        )
        print(message)
        if ignore_errors:
            print("-n (--ignore-errors) flag is set, continuing...\n")
            return
        raise ChildProcessError()


class TaskDone:
    def __init__(self, skipped=False) -> None:
        self.skipped = skipped


def initialize(path: Path) -> TaskDone:
    destination = path / "parameters.py"
    if destination.exists():
        return TaskDone(skipped=True)

    if not path.exists():
        path.mkdir(parents=True)
        print(f"created new directory: {path.absolute()}")

    root = get_project_root()
    template_path = root / "generate" / "parameters_template.py"

    # copy file
    destination.write_bytes(template_path.read_bytes())
    print(f"created template parameters file: {destination.absolute()}")

    return TaskDone()


def clean(args, path: Path) -> TaskDone:
    if not args.clean:
        return TaskDone(skipped=True)

    warn(f'--clean will delete all files in "{path}" except parameters.py')
    if not confirm("Are you sure?", default=False):
        return TaskDone()

    safe_to_delete = [
        r".*\.lammpstrj",
        r"log\.lammps",
        r"parameterfile",
        r"datafile.*",
        r"post_processing_parameters\.py",
        r"styles",
        r"tmp\.lammps\.variable",
        r"vmd\.tcl",
        r"states",
    ]
    safe_to_delete = [compile_regex(string) for string in safe_to_delete]

    def is_safe_to_delete(path: Path) -> bool:
        name = path.name
        return any(regex.match(name) for regex in safe_to_delete)

    def remove_recursively(path: Path):
        try:
            path.unlink()
        except IsADirectoryError:
            for child in path.iterdir():
                remove_recursively(child)
            path.rmdir()

    for child in path.iterdir():
        if child.name == "parameters.py":
            continue
        if not is_safe_to_delete(child):
            print(f"unrecognized file or folder '{child}', skipping...")
            continue

        remove_recursively(child)
        print(f"deleted '{child}' succesfully")

    return TaskDone()


def generate(args, path: Path) -> TaskDone:
    if not args.generate:
        if args.seed is not None:
            warn("seed argument is ignored when -g flag is not used!")
        return TaskDone(skipped=True)

    extra_args = []
    if args.seed:
        extra_args.append(args.seed)
    print("running setup file...")
    run_and_handle_error(
        lambda: subprocess.run(
            PYRUN + ["smc_lammps.generate.generate+parse", f"{path}"] + extra_args,
            check=False,
        ),
        args.ignore_errors,
    )
    print("successfully ran setup file")

    return TaskDone()


def get_lammps_args_list(lammps_vars: List[List[str]]):
    out = []
    for var in lammps_vars:
        out += ["-var"] + var
    return out


def perform_run(args, path: Path, lammps_vars: List[List[str]] | None = None):
    if lammps_vars is None:
        lammps_vars = []

    if args.input is None:
        project_root = get_project_root()
        args.input = project_root / "lammps" / "input.lmp"

    lammps_script = Path(args.input)
    lammps_vars.append(["lammps_root_dir", f"{lammps_script.parent.absolute()}"])

    command = [
        f"{args.executable}",
        "-sf",
        f"{args.suffix}",
        "-in",
        f"{lammps_script.absolute()}",
    ] + get_lammps_args_list(lammps_vars)
    if args.suffix == "kk":
        command += ["-kokkos", "on"]

    run_with_output = partial(subprocess.run, command, cwd=path.absolute())

    if args.output:
        with open(args.output, "w", encoding="utf-8") as output_file:
            print(
                f"running LAMMPS file {args.input}, output redirected to {args.output}"
            )
            print(command)
            run_and_handle_error(
                lambda: run_with_output(stdout=output_file), args.ignore_errors
            )
    else:
        print(f"running LAMMPS file {args.input}, printing output to terminal")
        print(command)
        run_and_handle_error(run_with_output, args.ignore_errors)


def restart_run(args, path: Path, output_file: Path) -> TaskDone:
    if not args.continue_flag:
        return TaskDone(skipped=True)

    file_exists = output_file.exists()
    if not file_exists:
        if args.force:
            return TaskDone(skipped=True)
        raise FileNotFoundError(
            "Make sure the following file exists to restart a simulation:", output_file
        )

    perform_run(args, path, [["is_restart", "1"]])

    return TaskDone()


def run(args, path: Path) -> TaskDone:
    if not args.run:
        return TaskDone(skipped=True)

    # check if output.lammpstrj exists
    output_file = path / "output.lammpstrj"

    if not restart_run(args, path, output_file).skipped:
        return TaskDone()

    if args.force:
        output_file.unlink(missing_ok=True)
        (path / "restartfile").unlink(missing_ok=True)
        (path / "perspective.output.lammpstrj").unlink(missing_ok=True)

    if output_file.exists():
        warn(
            "cannot run lammps script, output.lammpstrj already exists (use -f to overwrite files)"
        )
        print("moving on...")
        return TaskDone()

    perform_run(args, path)

    return TaskDone()


def post_process(args, path: Path) -> TaskDone:
    if not args.post_process:
        return TaskDone(skipped=True)

    print("running post processing...")
    run_and_handle_error(
        lambda: subprocess.run(
            PYRUN + ["smc_lammps.post-process.process_displacement", f"{path}"],
            check=False,
        ),
        args.ignore_errors,
    )
    print("succesfully ran post processing")

    return TaskDone()


def visualize_datafile(args, path: Path) -> TaskDone:
    if not args.visualize_datafile:
        return TaskDone(skipped=True)

    print("starting VMD")
    run_and_handle_error(
        lambda: subprocess.run(["vmd", "-e", f"{path}/vmd.tcl"], check=False),
        args.ignore_errors,
    )
    print("VMD exited")

    return TaskDone()


def create_perspective_file(args, path: Path, force=False):
    if not force and (path / "perspective.output.lammpstrj").exists():
        print("found perspective.output.lammpstrj")
        return

    print("creating new lammpstrj file")
    run_and_handle_error(
        lambda: subprocess.run(
            PYRUN
            + [
                "smc_lammps.post-process.smc_perspective",
                f"{path / 'output.lammpstrj'}",
                f"{path / 'post_processing_parameters.py'}",
            ],
            check=False,
        ),
        args.ignore_errors,
    )
    print("created perspective.output.lammpstrj")


def visualize_follow(args, path: Path) -> TaskDone:
    if not args.visualize_follow:
        return TaskDone(skipped=True)

    create_perspective_file(args, path)

    print("starting VMD")
    run_and_handle_error(
        lambda: subprocess.run(
            PYRUN
            + [
                "smc_lammps.post-process.visualize",
                f"{path}",
                "--file_name",
                "perspective.output.lammpstrj",
            ],
            check=False,
        ),
        args.ignore_errors,
    )
    print("VMD exited")

    return TaskDone()


def visualize(args, path: Path) -> TaskDone:
    if not visualize_datafile(args, path).skipped:
        return TaskDone()

    if not visualize_follow(args, path).skipped:
        return TaskDone()

    if not args.visualize:
        return TaskDone(skipped=True)

    print("starting VMD")
    run_and_handle_error(
        lambda: subprocess.run(
            PYRUN + ["smc_lammps.post-process.visualize", f"{path}"], check=False
        ),
        args.ignore_errors,
    )
    print("VMD exited")

    return TaskDone()


def main():
    args = parse()
    path = Path(args.directory)

    # --continue flag implies the --run flag
    if args.continue_flag:
        args.run = True

    tasks = [
        initialize(path),
        clean(args, path),
        generate(args, path),
        run(args, path),
        post_process(args, path),
        visualize(args, path),
    ]

    if all(map(lambda task: task.skipped, tasks)):
        print("nothing to do, use -gr to generate and run")

    print("end of run.py")


if __name__ == "__main__":
    # set PYTHONUNBUFFERED=1 if python is not printing correctly
    main()



from typing import List

import numpy as np
import numpy.typing as npt
from scipy import interpolate
from scipy.stats import trapezoid
from scipy.spatial.transform import Rotation

from smc_lammps.generate.generator import Nx3Array


def get_interpolated(spacing: float, values) -> Nx3Array:
    """spacing: distance between points along curve
    values: list of 3d points to use in the interpolation
    returns n equidistant points on an interpolated curve"""
    tck, u = interpolate.splprep(values.transpose())
    mi, ma = min(u), max(u)

    # calculate the length integral at many points
    sampling = np.linspace(mi, ma, 10000)
    derivatives_along_curve = np.array(
        interpolate.splev(sampling, tck, der=1)
    ).transpose()
    integrands = np.sqrt(np.sum(derivatives_along_curve**2, axis=1))
    lengths = np.array(
        [
            trapezoid(integrands[: i + 1], x=sampling[: i + 1])
            for i in range(len(integrands))
        ]
    )

    equidistant_points = [values[0]]
    while True:
        try:
            lengths -= spacing
            index = np.where(lengths >= 0)[0][0]
        except IndexError:
            break
        equidistant_points.append(interpolate.splev(sampling[index], tck))

    return np.array(equidistant_points)


# print(get_interpolated(100, np.array([[0, 0, 0], [1, 0, 0], [0.5, 0.5, 0], [0, 1, 0], [-4, 1, 0]])))


def get_straight_segment(n: int, direction=(1, 0, 0)) -> Nx3Array:
    """returns a straight segment of n beads with unit spacing starting at
    the origin and going the in provided direction (positive x-axis by default)"""
    direction = np.array(direction, dtype=np.float32)
    length: np.float32 = np.linalg.norm(direction)
    normalized_direction: npt.NDArray[np.float32] = direction / length
    segment = np.repeat(normalized_direction, n).reshape(3, n) * np.arange(n)
    return segment.transpose()


def get_circle_segment_unit_radius(
    n: int,
    end_inclusive: bool,
    theta_start: float = 0,
    theta_end: float = 2 * np.pi,
    normal_direction=(0, 0, 1),
) -> Nx3Array:
    normal_direction = np.array(normal_direction, dtype=float)

    arange = np.arange(n) / (n - 1 if end_inclusive else n)
    thetas = theta_start + arange * (theta_end - theta_start)
    segment = np.array([np.cos(thetas), np.sin(thetas), np.zeros(len(thetas))]).reshape(
        3, n
    )

    normal_direction /= np.linalg.norm(normal_direction)
    xy_normal = np.array([0, 0, 1], dtype=float)

    if np.linalg.norm(normal_direction - xy_normal) > 10 ** (-13):
        rotation_vector = np.cross(xy_normal, normal_direction)
        rotation_angle = np.arcsin(np.linalg.norm(rotation_vector))
        rotation = Rotation.from_rotvec(
            rotation_vector / np.linalg.norm(rotation_vector) * rotation_angle
        )
        segment = rotation.as_matrix().dot(segment)

    return segment.transpose()


def get_circle_segment(
    n: int,
    end_inclusive: bool,
    theta_start: float = 0,
    theta_end: float = 2 * np.pi,
    normal_direction=(0, 0, 1),
) -> Nx3Array:
    """returns a segment of a circle of n beads with unit spacing centered at the origin
    within the plane perpendical to the given normal_direction (in the x-y plane by default)"""
    segment = get_circle_segment_unit_radius(
        n, end_inclusive, theta_start, theta_end, normal_direction
    )
    if n < 2:
        return segment
    distance: np.float32 = np.linalg.norm(segment[0] - segment[1])
    return segment / distance


def attach(
    reference_segment: Nx3Array,
    other_segment: Nx3Array,
    delete_overlap: bool,
    extra_distance: float = 0.0,
) -> Nx3Array:
    """attaches the other_segment by moving its beginning to the end of the reference_segment"""
    extra_vector = np.zeros(len(reference_segment[0]))
    if isinstance(extra_distance, float):
        if extra_distance != 0.0:
            average_vector = (
                reference_segment[-1]
                - reference_segment[-2]
                + other_segment[1]
                - other_segment[0]
            ) / 2.0
            extra_vector = extra_distance * average_vector

    other_segment += reference_segment[-1] - other_segment[0] + extra_vector
    if delete_overlap:
        other_segment = other_segment[1:]

    return other_segment


def attach_chain(reference_segment: Nx3Array, list_of_args) -> List[Nx3Array]:
    """returns a list of the updated segments"""
    first_segment = reference_segment
    for i in range(len(list_of_args)):
        list_of_args[i][0] = attach(
            reference_segment, list_of_args[i][0], *list_of_args[i][1:]
        )
        reference_segment = list_of_args[i][0]
    return [first_segment] + [args[0] for args in list_of_args]
